""":mod:`wand.color` --- Colors
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. versionadded:: 0.1.2

"""
import ctypes
import numbers

from .api import library
from .cdefs.structures import MagickPixelPacket, PixelInfo
from .compat import binary, text
from .resource import Resource
from .version import MAGICK_VERSION_NUMBER, MAGICK_HDRI, QUANTUM_DEPTH

__all__ = 'Color', 'scale_quantum_to_int8'


class Color(Resource):
    """Color value.

    Unlike any other objects in Wand, its resource management can be
    implicit when it used outside of :keyword:`with` block. In these case,
    its resource are allocated for every operation which requires a resource
    and destroyed immediately. Of course it is inefficient when the
    operations are much, so to avoid it, you should use color objects
    inside of :keyword:`with` block explicitly e.g.::

        red_count = 0
        with Color('#f00') as red:
            with Image(filename='image.png') as img:
                for row in img:
                    for col in row:
                        if col == red:
                            red_count += 1

    :param string: a color name string e.g. ``'rgb(255, 255, 255)'``,
                   ``'#fff'``, ``'white'``. see `ImageMagick Color Names`_
                   doc also
    :type string: :class:`basestring`

    .. versionchanged:: 0.3.0
       :class:`Color` objects become hashable.

    .. versionchanged:: 0.5.1
       Color channel properties can now be set.

    .. versionchanged:: 0.5.1
       Added :attr:`cyan`, :attr:`magenta`, :attr:`yellow`, & :attr:`black`
       properties for CMYK :class:`Color` instances.

    .. versionchanged:: 0.5.1
       Method :meth:`Color.from_hsl()` can create a RGB color from ``hue``,
       ``saturation``, & ``lightness`` values.

    .. seealso::

       `ImageMagick Color Names`_
          The color can then be given as a color name (there is a limited
          but large set of these; see below) or it can be given as a set
          of numbers (in decimal or hexadecimal), each corresponding to
          a channel in an RGB or RGBA color model. HSL, HSLA, HSB, HSBA,
          CMYK, or CMYKA color models may also be specified. These topics
          are briefly described in the sections below.

    .. _ImageMagick Color Names: http://www.imagemagick.org/script/color.php

    .. describe:: == (other)

       Equality operator.

       :param other: a color another one
       :type color: :class:`Color`
       :returns: ``True`` only if two images equal.
       :rtype: :class:`bool`

    """

    #: (:class:`bool`) Whether the color has changed or not.
    dirty = None

    c_is_resource = library.IsPixelWand
    c_destroy_resource = library.DestroyPixelWand
    c_get_exception = library.PixelGetException
    c_clear_exception = library.PixelClearException

    __slots__ = 'raw', 'c_resource', 'allocated'

    def __init__(self, string=None, raw=None):
        if (string is None and raw is None or
                string is not None and raw is not None):
            raise TypeError('expected one argument')

        # MagickPixelPacket has been deprecated, use PixelInfo
        self.use_pixel = MAGICK_VERSION_NUMBER >= 0x700
        self.dirty = False
        self.allocated = 0
        if raw is None:
            if self.use_pixel:  # pragma: no cover
                self.raw = ctypes.create_string_buffer(
                    ctypes.sizeof(PixelInfo)
                )
            else:
                self.raw = ctypes.create_string_buffer(
                    ctypes.sizeof(MagickPixelPacket)
                )
            with self:
                # Create color from string.
                ok = library.PixelSetColor(self.resource, binary(string))
                if not ok:
                    # Could not understand color-input. Try sending
                    # ImageMagick's exception.
                    self.raise_exception()
                    # That might be only a warning. Try a more generic message.
                    msg = 'Unrecognized color string "{0}"'.format(string)
                    raise ValueError(msg)
                # Copy color value to structure buffer for future read.
                library.PixelGetMagickColor(self.resource, self.raw)
        else:
            self.raw = raw

    def __getinitargs__(self):
        return self.string, None

    def __enter__(self):
        if self.allocated < 1:
            with self.allocate():
                # Initialize resource.
                self.resource = library.NewPixelWand()
                # Restore color value from structure buffer.
                if self.use_pixel:  # pragma: no cover
                    library.PixelSetPixelColor(self.resource, self.raw)
                else:
                    library.PixelSetMagickColor(self.resource, self.raw)
            self.allocated = 1
        else:
            self.allocated += 1
        return Resource.__enter__(self)

    def __exit__(self, type, value, traceback):
        self.allocated -= 1
        if self.dirty:
            library.PixelGetMagickColor(self.resource, self.raw)
            self.dirty = False
        if self.allocated < 1:
            Resource.__exit__(self, type, value, traceback)

    def __eq__(self, other):
        if not isinstance(other, Color):
            return False
        with self as this:
            with other:
                return self.c_equals(this.resource, other.resource)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        if self.alpha:
            return hash(self.normalized_string)
        return hash(None)

    def __str__(self):
        return self.string

    def __repr__(self):
        c = type(self)
        return '{0}.{1}({2!r})'.format(c.__module__, c.__name__, self.string)

    def _assert_double(self, subject):
        """Ensure the given ``subject`` is a float type, and value between
        0.0 & 1.0.

        :param subject: value to assert as a valid double.
        :type subject: :class:`numbers.Real`
        :raises ValueError: if the subject is not between 0.0 and 1.0
        :raises TypeError: if the subject is not a float-point number.

        ..versionadded:: 0.5.1
        """
        if not isinstance(subject, numbers.Real):
            raise TypeError('Expecting a float-point real number, not ' +
                            repr(subject))
        if subject < 0.0 or subject > 1.0:
            raise ValueError('Expecting a real number between 0.0 & 1.0, not' +
                             repr(subject))

    def _assert_int8(self, subject):
        """Ensure the given ``subject`` is a integer type, and value between
        0 & 255.

        :param subject: value to assert as a valid number.
        :type subject: :class:`numbers.Integral`
        :raises ValueError: if the subject is not between 0 and 255
        :raises TypeError: if the subject is not a Integral number.

        ..versionadded:: 0.5.1
        """
        if not isinstance(subject, numbers.Integral):
            raise TypeError('Expecting an integer number, not ' +
                            repr(subject))
        if subject < 0 or subject > 255:
            raise ValueError('Expecting a real number between 0 & 255, not' +
                             repr(subject))

    def _assert_quantum(self, subject):
        """Ensure the given ``subject`` is a number, and value between
        0.0 & QuantumRange.

        The QuantumRange is the max value based on the QuantumDepth of the
        ImageMagick library (i.e. Q16).

        :param subject: value to assert as a valid double.
        :type subject: :class:`numbers.Number`
        :raises ValueError: if the subject is not between 0 and QuantumRange
        :raises TypeError: if the subject is not a number.

        ..versionadded:: 0.5.1
        """
        quantum_range = {
            8: 255.0,
            16: 65535.0,
            32: 4294967295.0,
            64: 18446744073709551615.0
        }
        if not isinstance(subject, numbers.Number):
            raise TypeError('Expecting a number, not ' + repr(subject))
        if subject < 0.0 or subject > quantum_range[QUANTUM_DEPTH]:
            message = 'Expecting a number between 0 & {0}, not {1}'
            raise ValueError(message.format(quantum_range[QUANTUM_DEPTH],
                                            repr(subject)))

    def _repr_html_(self):
        html = """
        <span style="background-color:#{red:02X}{green:02X}{blue:02X};
                     display:inline-block;
                     line-height:1em;
                     width:1em;">&nbsp;</span>
        <strong>#{red:02X}{green:02X}{blue:02X}</strong>
        """
        return html.format(red=self.red_int8,
                           green=self.green_int8,
                           blue=self.blue_int8)

    @staticmethod
    def c_equals(a, b):
        """Raw level version of equality test function for two pixels.

        :param a: a pointer to PixelWand to compare
        :type a: :class:`ctypes.c_void_p`
        :param b: a pointer to PixelWand to compare
        :type b: :class:`ctypes.c_void_p`
        :returns: ``True`` only if two pixels equal
        :rtype: :class:`bool`

        .. note::

           It's only for internal use. Don't use it directly.
           Use ``==`` operator of :class:`Color` instead.

        """
        alpha = library.PixelGetAlpha
        return bool(library.IsPixelWandSimilar(a, b, 0) and
                    alpha(a) == alpha(b))

    @classmethod
    def from_hsl(cls, hue=0.0, saturation=0.0, lightness=0.0):
        """Creates a RGB color from HSL values. The ``hue``, ``saturation``,
        and ``lightness`` must be normalized between 0.0 & 1.0.

        .. code::

            h=0.75  # 270 Degrees
            s=1.0   # 100 Percent
            l=0.5   # 50 Percent
            with Color.from_hsl(hue=h, saturation=s, lightness=l) as color:
                print(color)  #=> srgb(128,0,255)

        :param hue: a normalized double between 0.0 & 1.0.
        :type hue: :class:`numbers.Real`
        :param saturation: a normalized double between 0.0 & 1.0.
        :type saturation: :class:`numbers.Real`
        :param lightness: a normalized double between 0.0 & 1.0.
        :type lightness: :class:`numbers.Real`
        :rtype: :class:`Color`

        .. versionadded:: 0.5.1
        """
        color = cls('WHITE')
        color._assert_double(hue)
        color._assert_double(saturation)
        color._assert_double(lightness)
        color.dirty = True
        with color:
            library.PixelSetHSL(color.resource, hue, saturation, lightness)
        return color

    @classmethod
    def from_pixelwand(cls, pixelwand):
        assert pixelwand
        if MAGICK_VERSION_NUMBER < 0x700:
            pixel_structure = MagickPixelPacket
        else:  # pragma: no cover
            pixel_structure = PixelInfo
        size = ctypes.sizeof(pixel_structure)
        raw_buffer = ctypes.create_string_buffer(size)
        library.PixelGetMagickColor(pixelwand, raw_buffer)
        return cls(raw=raw_buffer)

    @property
    def alpha(self):
        """(:class:`numbers.Real`) Alpha value, from 0.0 to 1.0."""
        with self:
            return library.PixelGetAlpha(self.resource)

    @alpha.setter
    def alpha(self, value):
        self._assert_double(value)
        self.dirty = True
        with self:
            library.PixelSetAlpha(self.resource, value)

    @property
    def alpha_int8(self):
        """(:class:`numbers.Integral`) Alpha value as 8bit integer which is
        a common style.  From 0 to 255.

        .. versionadded:: 0.3.0

        """
        return scale_quantum_to_int8(self.alpha_quantum)

    @alpha_int8.setter
    def alpha_int8(self, value):
        self._assert_int8(value)
        self.alpha = float(value) / 255.0

    @property
    def alpha_quantum(self):
        """(:class:`numbers.Integral`) Alpha value.
        Scale depends on :const:`~wand.version.QUANTUM_DEPTH`.

        .. versionadded:: 0.3.0

        """
        with self:
            return library.PixelGetAlphaQuantum(self.resource)

    @alpha_quantum.setter
    def alpha_quantum(self, value):
        self._assert_quantum(value)
        self.dirty = True
        with self:
            library.PixelSetAlphaQuantum(self.resource, value)

    @property
    def black(self):
        """(:class:`numbers.Real`) Black, or ``'K'``, color channel in CMYK
        colorspace. Unused by RGB colorspace.

        .. versionadded:: 0.5.1
        """
        with self:
            return library.PixelGetBlack(self.resource)

    @black.setter
    def black(self, value):
        self._assert_double(value)
        self.dirty = True
        with self:
            library.PixelSetBlack(self.resource, value)

    @property
    def black_int8(self):
        """(:class:`numbers.Integral`) Black value as 8bit integer which is
        a common style.  From 0 to 255.

        .. versionadded:: 0.5.1
        """
        return scale_quantum_to_int8(self.black_quantum)

    @black_int8.setter
    def black_int8(self, value):
        self._assert_int8(value)
        self.black = float(value) / 255.0

    @property
    def black_quantum(self):
        """(:class:`numbers.Integral`) Black.
        Scale depends on :const:`~wand.version.QUANTUM_DEPTH`.

        .. versionadded:: 0.5.1
        """
        with self:
            return library.PixelGetBlackQuantum(self.resource)

    @black_quantum.setter
    def black_quantum(self, value):
        self._assert_quantum(value)
        self.dirty = True
        with self:
            library.PixelSetBlackQuantum(self.resource, value)

    @property
    def blue(self):
        """(:class:`numbers.Real`) Blue, from 0.0 to 1.0."""
        with self:
            return library.PixelGetBlue(self.resource)

    @blue.setter
    def blue(self, value):
        self._assert_double(value)
        self.dirty = True
        with self:
            library.PixelSetBlue(self.resource, value)

    @property
    def blue_int8(self):
        """(:class:`numbers.Integral`) Blue as 8bit integer which is
        a common style.  From 0 to 255.

        .. versionadded:: 0.3.0

        """
        return scale_quantum_to_int8(self.blue_quantum)

    @blue_int8.setter
    def blue_int8(self, value):
        self._assert_int8(value)
        self.blue = float(value) / 255.0

    @property
    def blue_quantum(self):
        """(:class:`numbers.Integral`) Blue.
        Scale depends on :const:`~wand.version.QUANTUM_DEPTH`.

        .. versionadded:: 0.3.0

        """
        with self:
            return library.PixelGetBlueQuantum(self.resource)

    @blue_quantum.setter
    def blue_quantum(self, value):
        self._assert_quantum(value)
        self.dirty = True
        with self:
            library.PixelSetBlueQuantum(self.resource, value)

    @property
    def cyan(self):
        """(:class:`numbers.Real`) Cyan color channel in CMYK
        colorspace. Unused by RGB colorspace.

        .. versionadded:: 0.5.1
        """
        with self:
            return library.PixelGetCyan(self.resource)

    @cyan.setter
    def cyan(self, value):
        self._assert_double(value)
        self.dirty = True
        with self:
            library.PixelSetCyan(self.resource, value)

    @property
    def cyan_int8(self):
        """(:class:`numbers.Integral`) Cyan value as 8bit integer which is
        a common style.  From 0 to 255.

        .. versionadded:: 0.5.1
        """
        return scale_quantum_to_int8(self.cyan_quantum)

    @cyan_int8.setter
    def cyan_int8(self, value):
        self._assert_int8(value)
        self.cyan = float(value) / 255.0

    @property
    def cyan_quantum(self):
        """(:class:`numbers.Integral`) Cyan.
        Scale depends on :const:`~wand.version.QUANTUM_DEPTH`.

        .. versionadded:: 0.5.1
        """
        with self:
            return library.PixelGetCyanQuantum(self.resource)

    @cyan_quantum.setter
    def cyan_quantum(self, value):
        self._assert_quantum(value)
        self.dirty = True
        with self:
            library.PixelSetCyanQuantum(self.resource, value)

    @property
    def fuzz(self):
        with self:
            return library.PixelGetFuzz(self.resource)

    @fuzz.setter
    def fuzz(self, value):
        if not isinstance(value, numbers.Real):
            raise TypeError('Expecting a float-point real number, not ' +
                            repr(value))
        self.dirty = True
        with self:
            library.PixelSetFuzz(self.resource, value)

    @property
    def green(self):
        """(:class:`numbers.Real`) Green, from 0.0 to 1.0."""
        with self:
            return library.PixelGetGreen(self.resource)

    @green.setter
    def green(self, value):
        self._assert_double(value)
        self.dirty = True
        with self:
            library.PixelSetGreen(self.resource, value)

    @property
    def green_int8(self):
        """(:class:`numbers.Integral`) Green as 8bit integer which is
        a common style.  From 0 to 255.

        .. versionadded:: 0.3.0

        """
        return scale_quantum_to_int8(self.green_quantum)

    @green_int8.setter
    def green_int8(self, value):
        self._assert_int8(value)
        self.green = float(value) / 255.0

    @property
    def green_quantum(self):
        """(:class:`numbers.Integral`) Green.
        Scale depends on :const:`~wand.version.QUANTUM_DEPTH`.

        .. versionadded:: 0.3.0

        """
        with self:
            return library.PixelGetGreenQuantum(self.resource)

    @green_quantum.setter
    def green_quantum(self, value):
        self._assert_quantum(value)
        self.dirty = True
        with self:
            library.PixelSetGreenQuantum(self.resource, value)

    @property
    def magenta(self):
        """(:class:`numbers.Real`) Magenta color channel in CMYK
        colorspace. Unused by RGB colorspace.

        .. versionadded:: 0.5.1
        """
        with self:
            return library.PixelGetMagenta(self.resource)

    @magenta.setter
    def magenta(self, value):
        self._assert_double(value)
        self.dirty = True
        with self:
            library.PixelSetMagenta(self.resource, value)

    @property
    def magenta_int8(self):
        """(:class:`numbers.Integral`) Magenta value as 8bit integer which is
        a common style.  From 0 to 255.

        .. versionadded:: 0.5.1
        """
        return scale_quantum_to_int8(self.magenta_quantum)

    @magenta_int8.setter
    def magenta_int8(self, value):
        self._assert_int8(value)
        self.magenta = float(value) / 255.0

    @property
    def magenta_quantum(self):
        with self:
            return library.PixelGetMagentaQuantum(self.resource)

    @magenta_quantum.setter
    def magenta_quantum(self, value):
        """(:class:`numbers.Integral`) Magenta.
        Scale depends on :const:`~wand.version.QUANTUM_DEPTH`.

        .. versionadded:: 0.5.1
        """
        self._assert_quantum(value)
        self.dirty = True
        with self:
            library.PixelSetMagentaQuantum(self.resource, value)

    @property
    def normalized_string(self):
        """(:class:`basestring`) The normalized string representation of
        the color.  The same color is always represented to the same
        string.

        .. versionadded:: 0.3.0

        """
        with self:
            string = library.PixelGetColorAsNormalizedString(self.resource)
            return text(string.value)

    @property
    def red(self):
        """(:class:`numbers.Real`) Red, from 0.0 to 1.0."""
        with self:
            return library.PixelGetRed(self.resource)

    @red.setter
    def red(self, value):
        self._assert_double(value)
        self.dirty = True
        with self:
            library.PixelSetRed(self.resource, value)

    @property
    def red_int8(self):
        """(:class:`numbers.Integral`) Red as 8bit integer which is a common
        style.  From 0 to 255.

        .. versionadded:: 0.3.0

        """
        return scale_quantum_to_int8(self.red_quantum)

    @red_int8.setter
    def red_int8(self, value):
        self._assert_int8(value)
        self.red = float(value) / 255.0

    @property
    def red_quantum(self):
        """(:class:`numbers.Integral`) Red.
        Scale depends on :const:`~wand.version.QUANTUM_DEPTH`.

        .. versionadded:: 0.3.0

        """
        with self:
            return library.PixelGetRedQuantum(self.resource)

    @red_quantum.setter
    def red_quantum(self, value):
        self._assert_quantum(value)
        self.dirty = True
        with self:
            library.PixelSetRedQuantum(self.resource, value)

    @property
    def string(self):
        """(:class:`basestring`) The string representation of the color."""
        with self:
            color_string = library.PixelGetColorAsString(self.resource)
            return text(color_string.value)

    @property
    def yellow(self):
        """(:class:`numbers.Real`) Yellow color channel in CMYK
        colorspace. Unused by RGB colorspace.

        .. versionadded:: 0.5.1
        """
        with self:
            return library.PixelGetYellow(self.resource)

    @yellow.setter
    def yellow(self, value):
        self._assert_double(value)
        self.dirty = True
        with self:
            library.PixelSetYellow(self.resource, value)

    @property
    def yellow_int8(self):
        """(:class:`numbers.Integral`) Yellow as 8bit integer which is a common
        style. From 0 to 255.

        .. versionadded:: 0.5.1
        """
        return scale_quantum_to_int8(self.yellow_quantum)

    @yellow_int8.setter
    def yellow_int8(self, value):
        self._assert_int8(value)
        self.yellow = float(value) / 255.0

    @property
    def yellow_quantum(self):
        """(:class:`numbers.Integral`) Yellow.
        Scale depends on :const:`~wand.version.QUANTUM_DEPTH`.

        .. versionadded:: 0.5.1
        """
        with self:
            return library.PixelGetYellowQuantum(self.resource)

    @yellow_quantum.setter
    def yellow_quantum(self, value):
        self._assert_quantum(value)
        self.dirty = True
        with self:
            library.PixelSetYellowQuantum(self.resource, value)

    def hsl(self):
        """Calculate the HSL color values from the RGB color.

        :returns: Tuple containing three normalized doubles, between 0.0 &
                  1.0, representing ``hue``, ``saturation``, and ``lightness``.
        :rtype: :class:`collections.Sequence`

        .. versionadded:: 0.5.1
        """
        hue = ctypes.c_double(0.0)
        saturation = ctypes.c_double(0.0)
        lightness = ctypes.c_double(0.0)
        with self:
            library.PixelGetHSL(self.resource,
                                ctypes.byref(hue),
                                ctypes.byref(saturation),
                                ctypes.byref(lightness))
        return (hue.value, saturation.value, lightness.value)


def scale_quantum_to_int8(quantum):
    """Straightforward port of :c:func:`ScaleQuantumToChar()` inline
    function.

    :param quantum: quantum value
    :type quantum: :class:`numbers.Integral`
    :returns: 8bit integer of the given ``quantum`` value
    :rtype: :class:`numbers.Integral`

    .. versionadded:: 0.3.0
    .. versionchanged:: 0.5.0
        Added HDRI support
    """
    if quantum <= 0:
        return 0
    table = {8: 1, 16: 257.0, 32: 16843009.0, 64: 72340172838076673.0}
    if MAGICK_HDRI:  # pragma: no cover
        if QUANTUM_DEPTH == 8:
            v = quantum / table[QUANTUM_DEPTH]
        elif QUANTUM_DEPTH == 16:
            v = ((int(quantum + 128) - (int(quantum + 128) >> 8)) >> 8)
        elif QUANTUM_DEPTH == 32:
            v = ((quantum + 8421504) / table[QUANTUM_DEPTH])
        elif QUANTUM_DEPTH == 64:
            v = quantum / table[QUANTUM_DEPTH]
    else:
        v = quantum / table[QUANTUM_DEPTH]
    if v >= 255:
        return 255
    return int(v + 0.5)



""":mod:`wand.cdefs.structures` --- MagickWand C typedefs
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. versionadded:: 0.5.0
"""
import ctypes
import os
import platform
import sys

__all__ = ('c_magick_char_p', 'c_magick_real_t', 'c_magick_size_t',
           'c_ssize_t')


class c_magick_char_p(ctypes.c_char_p):
    """This subclass prevents the automatic conversion behavior of
    :class:`ctypes.c_char_p`, allowing memory to be properly freed in the
    destructor.  It must only be used for non-const character pointers
    returned by ImageMagick functions.

    """

    def __del__(self):
        """Relinquishes memory allocated by ImageMagick.
        We don't need to worry about checking for ``NULL`` because
        :c:func:`MagickRelinquishMemory` does that for us.
        Note alslo that :class:`ctypes.c_char_p` has no
        :meth:`~object.__del__` method, so we don't need to
        (and indeed can't) call the superclass destructor.

        """
        try:
            from wand.api import library  # Lazy load global library
            library.MagickRelinquishMemory(self)
        except ImportError:
            # Python my be shutting down; and such, ``sys.meta_path``
            # may not be available.
            pass


if not hasattr(ctypes, 'c_ssize_t'):
    if ctypes.sizeof(ctypes.c_uint) == ctypes.sizeof(ctypes.c_void_p):
        ctypes.c_ssize_t = ctypes.c_int
    elif ctypes.sizeof(ctypes.c_ulong) == ctypes.sizeof(ctypes.c_void_p):
        ctypes.c_ssize_t = ctypes.c_long
    elif ctypes.sizeof(ctypes.c_ulonglong) == ctypes.sizeof(ctypes.c_void_p):
        ctypes.c_ssize_t = ctypes.c_longlong
c_ssize_t = ctypes.c_ssize_t


env_real = os.getenv('WAND_REAL_TYPE', 'auto')
if env_real in ('double', 'c_double'):
    c_magick_real_t = ctypes.c_double
elif env_real in ('longdouble', 'c_longdouble'):
    c_magick_real_t = ctypes.c_longdouble
else:
    # Attempt to guess MagickRealType size
    if sys.maxsize > 2**32:
        c_magick_real_t = ctypes.c_double
    else:
        c_magick_real_t = ctypes.c_longdouble
del env_real


# FIXME: Might need to rewrite to check against c_void_p size;
# like `c_ssize_t` above, and not against window platform.
if sys.maxsize > 2**32:
    c_magick_size_t = ctypes.c_size_t
elif platform.system() == "Windows":
    c_magick_size_t = ctypes.c_ulonglong
else:
    c_magick_size_t = ctypes.c_size_t



import base64
import inspect
import json
import os
try:
    from urllib import parse as urllib, request as urllib2
except ImportError:
    import urllib
    import urllib2

from py.path import local
from pytest import fixture, hookimpl, mark

from wand.display import display as display_fn
from wand.image import Image
from wand.version import MAGICK_VERSION, VERSION


def pytest_addoption(parser):
    parser.addoption('--skip-slow', action='store_true',
                     help='Skip slow tests')
    parser.addoption('--imgur-client-id',
                     help='Imgur.com api client id.  Use imgur.com for '
                          'display() fixture if present.  Useful for '
                          'debugging on CI',
                     default=os.environ.get('IMGUR_CLIENT_ID'))
    parser.addoption('--skip-pdf', action='store_true',
                     help='Skip any test with PDF documents.')
    parser.addoption('--skip-fft', action='store_true',
                     help='Skip any test with Forward Fourier Transform.')
    parser.addoption('--no-pdf', action='store_true',
                     help='Alias to --skip-pdf.')


def pytest_collection_modifyitems(config, items):
    skip_slow = False
    skip_pdf = False
    skip_fft = False
    if config.getoption('--skip-slow'):
        skip_slow = mark.skip('skipped; --skip-slow option is used')
    if config.getoption('--skip-pdf'):
        skip_pdf = mark.skip('skipped; --skip-pdf option is used')
    if config.getoption('--skip-fft'):
        skip_fft = mark.skip('skipped; --skip-fft option is used')
    if config.getoption('--no-pdf'):
        skip_pdf = mark.skip('skipped; --skip-pdf option is used')
    for item in items:
        if skip_slow and 'slow' in item.keywords:
            item.add_marker(skip_slow)
        if skip_pdf and 'pdf' in item.keywords:
            item.add_marker(skip_pdf)
        if skip_fft and 'fft' in item.keywords:
            item.add_marker(skip_fft)


def pytest_configure(config):
    config.addinivalue_line(
        'markers', 'slow: marks test as slow-running'
    )
    config.addinivalue_line(
        'markers', 'pdf: marks test as PDF/Ghostscript dependent'
    )
    config.addinivalue_line(
        'markers', 'fft: marks test as Forward Fourier Transform dependent'
    )


@hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """Copied from http://pytest.org/dev/example/simple.html#making-test-result-information-available-in-fixtures

    """  # noqa
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"

    setattr(item, "rep_" + rep.when, rep)


def pytest_report_header(config):
    versions = (VERSION, os.linesep, MAGICK_VERSION)
    return "Wand Version: {0}{1}ImageMagick Version: {2}".format(*versions)


@fixture
def fx_asset():
    """The fixture that provides :class:`py.path.local` instance that
    points the :file:`assets` directory.  You can use this in test
    functions::

        def test_something(fx_asset):
            monalisa = str(fx_asset.join('mona-lisa.jpg'))
            with open(monalisa) as f:
                assert f.tell() == 0

    """
    return local(__file__).dirpath('assets')


@fixture
def display(request):
    """Display an given image when a test fails.  It's a function that
    takes one required argument of :class:`~wand.image.Image` and one
    optional argument of its human readable label.

    For example::

        def test_something(display):
            with Image(filename='...') as image:
                display(image)
                image.resize(100, 100)
                display(image)
                assert image.size == (100, 101)

    if the above test fails, it will display two images.

    If you give ``--imgur-client-id`` option to pytest, it uploads
    these images instead of displaying:

    .. code-block::

       test_something.py:3 (image 1) http://i.imgur.com/iJPHO68.png
       test_something.py:5 (image 2) http://i.imgur.com/iJPHO69.png

    This option would be useful for remote debugging on CI.

    """
    images = []

    @request.addfinalizer
    def finalize():
        if request.node.rep_call.passed:
            return
        imgur_client_id = request.config.getoption('--imgur-client-id')
        if imgur_client_id:
            longrepr = request.node.rep_call.longrepr
            prints = []
            for line, label, format, blob in images:
                req = urllib2.Request(
                    'https://api.imgur.com/3/image',
                    headers={'Authorization': 'Client-ID ' + imgur_client_id},
                    data=urllib.urlencode({
                        'image': base64.b64encode(blob),
                        'title': '[{0}] {1}'.format(line, label)
                    }).encode('ascii')
                )
                response = urllib2.urlopen(req)
                result = json.loads(response.read().decode('utf-8'))
                prints.append((line, label, result['data']['link']))
                response.close()
            longrepr.addsection(
                'display',
                '\n'.join(map('{0[0]} ({0[1]}) {0[2]}'.format, prints))
            )
        else:
            for _, __, format, blob in images:
                with Image(blob=blob, format=format) as i:
                    display_fn(i)

    def log(image, label=None):
        back = inspect.currentframe().f_back
        line = '{0}:{1}'.format(back.f_code.co_filename, back.f_lineno)
        label = label or 'image {0}'.format(len(images))
        format = image.format or 'png'
        images.append((
            line, label, format,
            image.make_blob(format)
        ))
    return log



""":mod:`wand.api` --- Low-level interfaces
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. versionchanged:: 0.1.10
   Changed to throw :exc:`~exceptions.ImportError` instead of
   :exc:`~exceptions.AttributeError` when the shared library fails to load.

"""
import ctypes
import ctypes.util
import itertools
import os
import os.path
import platform
import sys
import traceback
# Forward import for backwords compatibility.
from .cdefs.structures import (AffineMatrix, MagickPixelPacket, PixelInfo,
                               PointInfo)
if platform.system() == "Windows":
    try:
        import winreg
    except ImportError:
        import _winreg as winreg

__all__ = ('AffineMatrix', 'MagickPixelPacket', 'library', 'libc', 'libmagick',
           'load_library', 'PixelInfo', 'PointInfo')


def library_paths():
    """Iterates for library paths to try loading.  The result paths are not
    guaranteed that they exist.

    :returns: a pair of libwand and libmagick paths.  they can be the same.
              path can be ``None`` as well
    :rtype: :class:`tuple`

    """
    libwand = None
    libmagick = None
    versions = '', '-7', '-7.Q8', '-7.Q16', '-6', '-Q16', '-Q8', '-6.Q16'
    options = '', 'HDRI', 'HDRI-2'
    system = platform.system()
    magick_home = os.environ.get('MAGICK_HOME')
    magick_suffix = os.environ.get('WAND_MAGICK_LIBRARY_SUFFIX')

    if system == 'Windows':
        # ImageMagick installers normally install coder and filter DLLs in
        # subfolders, we need to add those folders to PATH, otherwise loading
        # the DLL later will fail.
        try:
            with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,
                                r"SOFTWARE\ImageMagick\Current") as reg_key:
                libPath = winreg.QueryValueEx(reg_key, "LibPath")
                coderPath = winreg.QueryValueEx(reg_key, "CoderModulesPath")
                filterPath = winreg.QueryValueEx(reg_key, "FilterModulesPath")
                magick_home = libPath[0]
                os.environ['PATH'] += str((';' + libPath[0] + ";" +
                                          coderPath[0] + ";" + filterPath[0]))
        except OSError:
            # otherwise use MAGICK_HOME, and we assume the coder and
            # filter DLLs are in the same directory
            pass

    def magick_path(path):
        return os.path.join(magick_home, *path)
    combinations = itertools.product(versions, options)
    suffixes = list()
    if magick_suffix:
        suffixes = str(magick_suffix).split(';')
    # We need to convert the ``combinations`` generator to a list so we can
    # iterate over it twice.
    suffixes.extend(list(version + option for version, option in combinations))
    if magick_home:
        # exhaustively search for libraries in magick_home before calling
        # find_library.
        for suffix in suffixes:
            # On Windows, the API is split between two libs. On other
            # platforms, it's all contained in one.
            if system == 'Windows':
                libwand = 'CORE_RL_wand_{0}.dll'.format(suffix),
                libmagick = 'CORE_RL_magick_{0}.dll'.format(suffix),
                yield magick_path(libwand), magick_path(libmagick)
                libwand = 'CORE_RL_MagickWand_{0}.dll'.format(suffix),
                libmagick = 'CORE_RL_MagickCore_{0}.dll'.format(suffix),
                yield magick_path(libwand), magick_path(libmagick)
                libwand = 'libMagickWand{0}.dll'.format(suffix),
                libmagick = 'libMagickCore{0}.dll'.format(suffix),
                yield magick_path(libwand), magick_path(libmagick)
            elif system == 'Darwin':
                libwand = 'lib', 'libMagickWand{0}.dylib'.format(suffix),
                yield magick_path(libwand), magick_path(libwand)
            else:
                libwand = 'lib', 'libMagickWand{0}.so'.format(suffix),
                libmagick = 'lib', 'libMagickCore{0}.so'.format(suffix),
                yield magick_path(libwand), magick_path(libmagick)
                libwand = 'lib', 'libMagickWand{0}.so.6'.format(suffix),
                libmagick = 'lib', 'libMagickCore{0}.so.6'.format(suffix),
                yield magick_path(libwand), magick_path(libmagick)
    for suffix in suffixes:
        if system == 'Windows':
            libwand = ctypes.util.find_library('CORE_RL_wand_' + suffix)
            libmagick = ctypes.util.find_library('CORE_RL_magick_' + suffix)
            yield libwand, libmagick
            libwand = ctypes.util.find_library('CORE_RL_MagickWand_' + suffix)
            libmagick = ctypes.util.find_library(
                'CORE_RL_MagickCore_' + suffix
            )
            yield libwand, libmagick
            libwand = ctypes.util.find_library('libMagickWand' + suffix)
            libmagick = ctypes.util.find_library('libMagickCore' + suffix)
            yield libwand, libmagick
        else:
            libwand = ctypes.util.find_library('MagickWand' + suffix)
            yield libwand, libwand


def load_library():
    """Loads the MagickWand library.

    :returns: the MagickWand library and the ImageMagick library
    :rtype: :class:`ctypes.CDLL`

    """
    tried_paths = []
    for libwand_path, libmagick_path in library_paths():
        if libwand_path is None or libmagick_path is None:
            continue
        try:
            tried_paths.append(libwand_path)
            libwand = ctypes.CDLL(str(libwand_path))
            if libwand_path == libmagick_path:
                libmagick = libwand
            else:
                tried_paths.append(libmagick_path)
                libmagick = ctypes.CDLL(str(libmagick_path))
        except (IOError, OSError):
            continue
        return libwand, libmagick
    raise IOError('cannot find library; tried paths: ' + repr(tried_paths))


try:
    # Preserve the module itself even if it fails to import
    sys.modules['wand._api'] = sys.modules['wand.api']
except KeyError:
    # Loading the module locally or a non-standard setting
    pass

try:
    libraries = load_library()
except (OSError, IOError):
    msg = 'http://docs.wand-py.org/en/latest/guide/install.html'
    if sys.platform.startswith(('dragonfly', 'freebsd')):
        msg = 'pkg install'
    elif sys.platform == 'win32':
        msg += '#install-imagemagick-on-windows'
    elif sys.platform == 'darwin':
        mac_pkgmgrs = {'brew': 'brew install freetype imagemagick',
                       'port': 'port install imagemagick'}
        for pkgmgr in mac_pkgmgrs:
            with os.popen('which ' + pkgmgr) as f:
                if f.read().strip():
                    msg = mac_pkgmgrs[pkgmgr]
                    break
        else:
            msg += '#install-imagemagick-on-mac'
    elif hasattr(platform, 'linux_distribution'):
        distname, _, __ = platform.linux_distribution()
        distname = (distname or '').lower()
        if distname in ('debian', 'ubuntu'):
            msg = 'apt-get install libmagickwand-dev'
        elif distname in ('fedora', 'centos', 'redhat'):
            msg = 'yum install ImageMagick-devel'
    raise ImportError('MagickWand shared library not found.\n'
                      'You probably had not installed ImageMagick library.\n'
                      'Try to install:\n  ' + msg)

#: (:class:`ctypes.CDLL`) The MagickWand library.
library = libraries[0]

#: (:class:`ctypes.CDLL`) The ImageMagick library.  It is the same with
#: :data:`library` on platforms other than Windows.
#:
#: .. versionadded:: 0.1.10
libmagick = libraries[1]

try:
    from wand.cdefs import (core, magick_wand, magick_image, magick_property,
                            pixel_iterator, pixel_wand, drawing_wand)

    core.load(libmagick)
    # Let's get the magick-version number to pass to load methods.
    IM_VERSION = ctypes.c_size_t()
    libmagick.GetMagickVersion(ctypes.byref(IM_VERSION))
    # Query Quantum Depth (i.e. Q8, Q16, ... etc).
    IM_QUANTUM_DEPTH = ctypes.c_size_t()
    libmagick.GetMagickQuantumDepth(ctypes.byref(IM_QUANTUM_DEPTH))
    # Does the library support HDRI?
    IM_HDRI = 'HDRI' in str(libmagick.GetMagickFeatures())
    core.load_with_version(libmagick, IM_VERSION.value)
    magick_wand.load(library, IM_VERSION.value)
    magick_property.load(library, IM_VERSION.value)
    magick_image.load(library, IM_VERSION.value)
    pixel_iterator.load(library, IM_VERSION.value)
    pixel_wand.load(library, IM_VERSION.value, IM_QUANTUM_DEPTH.value, IM_HDRI)
    drawing_wand.load(library, IM_VERSION.value)
    del IM_HDRI, IM_QUANTUM_DEPTH, IM_VERSION

except AttributeError:
    raise ImportError('MagickWand shared library not found or incompatible\n'
                      'Original exception was raised in:\n' +
                      traceback.format_exc())

#: (:class:`ctypes.CDLL`) The C standard library.
libc = None

if platform.system() == 'Windows':
    msvcrt = ctypes.util.find_msvcrt()
    # workaround -- the newest visual studio DLL is named differently:
    if not msvcrt and '1900' in platform.python_compiler():
        msvcrt = 'vcruntime140.dll'
    if msvcrt:
        libc = ctypes.CDLL(msvcrt)
else:
    libc_path = ctypes.util.find_library('c')
    if libc_path:
        libc = ctypes.cdll.LoadLibrary(libc_path)
    else:
        # Attempt to guess popular versions of libc
        libc_paths = ('libc.so.6', 'libc.so', 'libc.a', 'libc.dylib',
                      '/usr/lib/libc.dylib')
        for libc_path in libc_paths:
            try:
                libc = ctypes.cdll.LoadLibrary(libc_path)
                break
            except (IOError, OSError):
                continue
    if libc:
        libc.fdopen.argtypes = [ctypes.c_int, ctypes.c_char_p]
        libc.fdopen.restype = ctypes.c_void_p
        libc.fflush.argtypes = [ctypes.c_void_p]



""":mod:`wand.cdefs.pixel_wand` --- Pixel-Wand definitions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. versionadded:: 0.5.0
"""
from ctypes import (CDLL, POINTER, c_char_p, c_double,
                    c_float, c_int, c_longdouble, c_size_t,
                    c_ubyte, c_uint, c_ushort, c_void_p)
from wand.cdefs.wandtypes import c_magick_char_p
import numbers

__all__ = ('load',)


def load(lib, IM_VERSION, IM_QUANTUM_DEPTH, IM_HDRI):
    """Define Pixel Wand methods. The ImageMagick version is given as
    a second argument for comparison. This will quick to determine which
    methods are available from the library, and can be implemented as::

        if IM_VERSION < 0x700:
            # ... do ImageMagick-6 methods ...
        else
            # ... do ImageMagick-7 methods ...

    .. seealso::

        #include "wand/pixel-wand.h"
        // Or
        #include "MagickWand/pixel-wand.h"

    Mapping Pixel methods also requires the wand library to evaluate
    what "Quantum" is to ImageMagick. We must query the library
    to identify if HRDI is enabled, and what the quantum depth is.

    .. seealso::

        MagickCore/magick-type.h

    :param lib: the loaded ``MagickWand`` library.
    :type lib: :class:`ctypes.CDLL`
    :param IM_VERSION: the ImageMagick version number (i.e. 0x0689).
    :type IM_VERSION: :class:`numbers.Integral`
    :param IM_QUANTUM_DEPTH: the ImageMagick Quantum Depth
                             (must be 8, 16, 32, or 64).
    :type IM_QUANTUM_DEPTH: :class:`numbers.Integral`
    :param IM_HDRI: if ImageMagick was compiled with HDRI support.
    :type IM_HDRI: :class:`bool`

    .. versionadded:: 0.5.0

    """
    if not isinstance(lib, CDLL):
        raise AttributeError(repr(lib) + " is not an instanced of ctypes.CDLL")
    if not isinstance(IM_VERSION, numbers.Integral):
        raise AttributeError("Expecting MagickCore version number")
    if IM_QUANTUM_DEPTH not in [8, 16, 32, 65]:
        raise AttributeError("QUANTUM_DEPTH must be one of 8, 16, 32, or 64")
    is_im_6 = IM_VERSION < 0x700
    is_im_7 = IM_VERSION >= 0x700

    if IM_QUANTUM_DEPTH == 8:
        QuantumType = c_float if IM_HDRI else c_ubyte
    elif IM_QUANTUM_DEPTH == 16:
        QuantumType = c_float if IM_HDRI else c_ushort
    elif IM_QUANTUM_DEPTH == 32:
        QuantumType = c_double if IM_HDRI else c_uint
    elif IM_QUANTUM_DEPTH == 64:
        QuantumType = c_longdouble

    lib.ClearPixelWand.argtypes = [c_void_p]
    lib.ClonePixelWand.argtypes = [c_void_p]
    lib.ClonePixelWand.restype = c_void_p
    lib.DestroyPixelWand.argtypes = [c_void_p]
    lib.DestroyPixelWand.restype = c_void_p
    lib.DestroyPixelWands.argtypes = [POINTER(c_void_p), c_size_t]
    lib.DestroyPixelWands.restype = POINTER(c_void_p)
    lib.IsPixelWand.argtypes = [c_void_p]
    lib.IsPixelWandSimilar.argtypes = [c_void_p, c_void_p, c_double]
    lib.NewPixelWand.argtypes = []
    lib.NewPixelWand.restype = c_void_p
    lib.PixelClearException.argtypes = [c_void_p]
    lib.PixelClearException.restype = c_int
    lib.PixelGetAlpha.argtypes = [c_void_p]
    lib.PixelGetAlpha.restype = c_double
    lib.PixelGetAlphaQuantum.argtypes = [c_void_p]
    lib.PixelGetAlphaQuantum.restype = QuantumType
    lib.PixelGetBlack.argtypes = [c_void_p]
    lib.PixelGetBlack.restype = c_double
    lib.PixelGetBlackQuantum.argtypes = [c_void_p]
    lib.PixelGetBlackQuantum.restype = QuantumType
    lib.PixelGetBlue.argtypes = [c_void_p]
    lib.PixelGetBlue.restype = c_double
    lib.PixelGetBlueQuantum.argtypes = [c_void_p]
    lib.PixelGetBlueQuantum.restype = QuantumType
    lib.PixelGetColorAsNormalizedString.argtypes = [c_void_p]
    lib.PixelGetColorAsNormalizedString.restype = c_magick_char_p
    lib.PixelGetColorAsString.argtypes = [c_void_p]
    lib.PixelGetColorAsString.restype = c_magick_char_p
    lib.PixelGetColorCount.argtypes = [c_void_p]
    lib.PixelGetColorCount.restype = c_size_t
    lib.PixelGetCyan.argtypes = [c_void_p]
    lib.PixelGetCyan.restype = c_double
    lib.PixelGetCyanQuantum.argtypes = [c_void_p]
    lib.PixelGetCyanQuantum.restype = QuantumType
    lib.PixelGetException.argtypes = [c_void_p, POINTER(c_int)]
    lib.PixelGetException.restype = c_magick_char_p
    lib.PixelGetExceptionType.argtypes = [c_void_p]
    lib.PixelGetExceptionType.restype = c_int
    lib.PixelGetFuzz.argtypes = [c_void_p]
    lib.PixelGetFuzz.restype = c_double
    lib.PixelGetGreen.argtypes = [c_void_p]
    lib.PixelGetGreen.restype = c_double
    lib.PixelGetGreenQuantum.argtypes = [c_void_p]
    lib.PixelGetGreenQuantum.restype = QuantumType
    lib.PixelGetHSL.argtypes = [c_void_p,
                                POINTER(c_double),
                                POINTER(c_double),
                                POINTER(c_double)]
    lib.PixelGetIndex.argtypes = [c_void_p]
    lib.PixelGetIndex.restype = QuantumType
    lib.PixelGetMagenta.argtypes = [c_void_p]
    lib.PixelGetMagenta.restype = c_double
    lib.PixelGetMagentaQuantum.argtypes = [c_void_p]
    lib.PixelGetMagentaQuantum.restype = QuantumType
    lib.PixelGetMagickColor.argtypes = [c_void_p, c_void_p]
    if is_im_7:
        lib.PixelGetPixel.argtypes = [c_void_p]
        lib.PixelGetPixel.restype = c_void_p
    lib.PixelGetRed.argtypes = [c_void_p]
    lib.PixelGetRed.restype = c_double
    lib.PixelGetRedQuantum.argtypes = [c_void_p]
    lib.PixelGetRedQuantum.restype = QuantumType
    lib.PixelGetYellow.argtypes = [c_void_p]
    lib.PixelGetYellow.restype = c_double
    lib.PixelGetYellowQuantum.argtypes = [c_void_p]
    lib.PixelGetYellowQuantum.restype = QuantumType
    lib.PixelSetAlpha.argtypes = [c_void_p, c_double]
    lib.PixelSetAlphaQuantum.argtypes = [c_void_p, QuantumType]
    lib.PixelSetBlack.argtypes = [c_void_p, c_double]
    lib.PixelSetBlackQuantum.argtypes = [c_void_p, QuantumType]
    lib.PixelSetBlue.argtypes = [c_void_p, c_double]
    lib.PixelSetBlueQuantum.argtypes = [c_void_p, QuantumType]
    lib.PixelSetColor.argtypes = [c_void_p, c_char_p]
    lib.PixelSetColor.restype = c_int
    lib.PixelSetColorCount.argtypes = [c_void_p, c_size_t]
    lib.PixelSetCyan.argtypes = [c_void_p, c_double]
    lib.PixelSetCyanQuantum.argtypes = [c_void_p, QuantumType]
    lib.PixelSetFuzz.argtypes = [c_void_p, c_double]
    lib.PixelSetGreen.argtypes = [c_void_p, c_double]
    lib.PixelSetGreenQuantum.argtypes = [c_void_p, QuantumType]
    lib.PixelSetHSL.argtypes = [c_void_p, c_double, c_double, c_double]
    lib.PixelSetIndex.argtypes = [c_void_p, QuantumType]
    lib.PixelSetMagenta.argtypes = [c_void_p, c_double]
    lib.PixelSetMagentaQuantum.argtypes = [c_void_p, QuantumType]
    if is_im_6:
        lib.PixelSetMagickColor.argtypes = [c_void_p, c_void_p]
    else:
        lib.PixelSetMagickColor = None
    if is_im_7:
        lib.PixelSetPixelColor.argtypes = [c_void_p, c_void_p]
    else:
        lib.PixelSetPixelColor = None
    lib.PixelSetRed.argtypes = [c_void_p, c_double]
    lib.PixelSetRedQuantum.argtypes = [c_void_p, QuantumType]
    lib.PixelSetYellow.argtypes = [c_void_p, c_double]
    lib.PixelSetYellowQuantum.argtypes = [c_void_p, QuantumType]
    if is_im_6:
        lib.PixelSetMagickColor.argtypes = [c_void_p, c_void_p]
        lib.PixelSetPixelColor = None
    if is_im_7:
        lib.PixelSetMagickColor = None
        lib.PixelSetPixelColor.argtypes = [c_void_p, c_void_p]



# -*- coding: utf-8 -*-
#
# These test cover the Image attributes that directly map to C-API functions.
#
import io
import numbers
import warnings
from pytest import mark, raises

from wand.color import Color
from wand.compat import string_type
from wand.exceptions import DelegateError
from wand.font import Font
from wand.image import Image
from wand.version import MAGICK_VERSION_NUMBER


def test_alpha_channel_get(fx_asset):
    """Checks if image has alpha channel."""
    with Image(filename=str(fx_asset.join('watermark.png'))) as img:
        assert img.alpha_channel is True
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        assert img.alpha_channel is False


def test_alpha_channel_set(fx_asset):
    """Sets alpha channel to off."""
    with Image(filename=str(fx_asset.join('watermark.png'))) as img:
        if MAGICK_VERSION_NUMBER < 0x700:
            enable_option = 'on'
            disable_option = False
        else:
            enable_option = 'associate'
            disable_option = 'disassociate'
        img.alpha_channel = enable_option
        assert img.alpha_channel is True
        img.alpha_channel = disable_option
        assert img.alpha_channel is False
        img.alpha_channel = 'opaque'
        assert img[0, 0].alpha == 1.0
        with raises(ValueError):
            img.alpha_channel = 'watermark'


def test_artifacts():
    with Image(filename='rose:') as img:
        img.artifacts['key'] = 'value'
        assert 'date:create' in img.artifacts
        assert img.artifacts['key'] == 'value'
        assert img.artifacts['novalue'] is None
        assert len(img.artifacts) > 0
        for key, value in img.artifacts.items():
            pass
        del img.artifacts['key']


def test_background_color_get(fx_asset):
    """Gets the background color."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        assert Color('white') == img.background_color


def test_background_color_set(fx_asset):
    """Sets the background color."""
    with Image(filename=str(fx_asset.join('croptest.png'))) as img:
        with Color('red') as color:
            img.background_color = color
            assert img.background_color == color
        img.background_color = 'green'
        assert img.background_color == Color('green')


def test_border_color():
    green = Color('green')
    with Image(filename='rose:') as img:
        img.border_color = 'green'
        assert img.border_color == green


@mark.xfail(MAGICK_VERSION_NUMBER >= 0x700,
            reason="Channel traits are not implemented in IM7.")
def test_channel_depths(fx_asset):
    with Image(filename=str(fx_asset.join('beach.jpg'))) as i:
        assert dict(i.channel_depths) == {
            'blue': 8, 'gray': 8, 'true_alpha': 1, 'opacity': 1,
            'undefined': 1, 'composite_channels': 8, 'index': 1,
            'rgb_channels': 8, 'alpha': 1, 'yellow': 8, 'sync_channels': 1,
            'default_channels': 8, 'black': 1, 'cyan': 8,
            'all_channels': 8, 'green': 8, 'magenta': 8, 'red': 8,
            'gray_channels': 8, 'rgb': 8
        }
    with Image(filename=str(fx_asset.join('google.ico'))) as i:
        assert dict(i.channel_depths) == {
            'blue': 8, 'gray': 8, 'true_alpha': 1, 'opacity': 1,
            'undefined': 1, 'composite_channels': 8, 'index': 1,
            'rgb_channels': 8, 'alpha': 1, 'yellow': 8, 'sync_channels': 1,
            'default_channels': 8, 'black': 1, 'cyan': 8, 'all_channels': 8,
            'green': 8, 'magenta': 8, 'red': 8, 'gray_channels': 8, 'rgb': 8
        }


def test_channel_images(fx_asset):
    with Image(filename=str(fx_asset.join('sasha.jpg'))) as i:
        i.format = 'png'
        channels = ('opacity', 'alpha',)
        # Only include TrueAlphaChannel if IM6, as its deprecated & unused
        # in IM7.
        if MAGICK_VERSION_NUMBER < 0x700:
            channels = channels + ('true_alpha',)
        for name in channels:
            expected_path = str(fx_asset.join('channel_images', name + '.png'))
            with Image(filename=expected_path) as expected:
                if MAGICK_VERSION_NUMBER >= 0x700:
                    # With IM7, channels are dynamic & influence signatures.
                    # We'll need to compare the first channel of the expected
                    # PNG with the extracted channel.
                    first_channel = expected.channel_images['red']
                    assert i.channel_images[name] == first_channel
                else:
                    assert i.channel_images[name] == expected


def test_colors(fx_asset):
    with Image(filename=str(fx_asset.join('trim-color-test.png'))) as img:
        assert img.colors == 2


def test_colorspace_get(fx_asset):
    """Gets the image colorspace"""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        assert img.colorspace.endswith('rgb')


def test_colorspace_set(fx_asset):
    """Sets the image colorspace"""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        img.colorspace = 'cmyk'
        assert img.colorspace == 'cmyk'


def test_compose(fx_asset):
    with Image(filename=str(fx_asset.join('sasha.jpg'))) as img:
        assert img.compose == 'over'
        img.compose = 'blend'
        assert img.compose == 'blend'
        with raises(TypeError):
            img.compose = 0xDEADBEEF
        with raises(ValueError):
            img.compose = 'none'


def test_compression(fx_asset):
    with Image(filename=str(fx_asset.join('sasha.jpg'))) as img:
        # Legacy releases/library asserted ``'group4'`` compression type.
        # IM 7 will correctly report ``'jpeg'``, but ``'group4'`` should
        # still be apart of regression acceptance.
        assert img.compression in ('group4', 'jpeg')
        img.compression = 'zip'
        assert img.compression == 'zip'
        with raises(TypeError):
            img.compression = 0x60


def test_compression_quality_get(fx_asset):
    """Gets the image compression quality."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        assert img.compression_quality == 80


def test_compression_quality_set(fx_asset):
    """Sets the image compression quality."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        img.compression_quality = 50
        assert img.compression_quality == 50
        strio = io.BytesIO()
        img.save(file=strio)
        strio.seek(0)
        with Image(file=strio) as jpg:
            assert jpg.compression_quality == 50
        with raises(TypeError):
            img.compression_quality = 'high'


def test_delay_set_get(fx_asset):
    with Image(filename=str(fx_asset.join('nocomments.gif'))) as img:
        img.delay = 10
        assert img.delay == 10


def test_depth_get(fx_asset):
    """Gets the image depth"""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        assert img.depth == 8


def test_depth_set(fx_asset):
    """Sets the image depth"""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        img.depth = 16
        assert img.depth == 16


def test_dispose(fx_asset):
    with Image(filename=str(fx_asset.join('nocomments.gif'))) as img:
        assert img.dispose == 'none'
        img.dispose = 'background'
        assert img.dispose == 'background'


def test_font_set(fx_asset):
    with Image(width=144, height=192, background=Color('#1e50a2')) as img:
        font = Font(
            path=str(fx_asset.join('League_Gothic.otf')),
            color=Color('gold'),
            size=12,
            antialias=False
        )
        img.font = font
        assert img.font_path == font.path
        assert img.font_size == font.size
        assert img.font_color == font.color
        assert img.antialias == font.antialias
        assert img.font == font
        assert repr(img.font)
        fontStroke = Font(
            path=str(fx_asset.join('League_Gothic.otf')),
            stroke_color=Color('ORANGE'),
            stroke_width=1.5
        )
        img.font = fontStroke
        assert img.stroke_color == fontStroke.stroke_color
        assert img.stroke_width == fontStroke.stroke_width
        img.font_color = 'gold'
        assert img.font_color == Color('gold')
        img.stroke_color = 'gold'
        assert img.stroke_color == Color('gold')
        fontColor = Font(
            path=str(fx_asset.join('League_Gothic.otf')),
            color='YELLOW',
            stroke_color='PINK'
        )
        img.font = fontColor
        assert img.font_color == Color('YELLOW')
        assert img.stroke_color == Color('PINK')
        with raises(ValueError):
            img.font_size = -99


def test_format_get(fx_asset):
    """Gets the image format."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        assert img.format == 'JPEG'
    with Image(filename=str(fx_asset.join('croptest.png'))) as img:
        assert img.format == 'PNG'


def test_format_set(fx_asset):
    """Sets the image format."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        img.format = 'png'
        assert img.format == 'PNG'
        strio = io.BytesIO()
        img.save(file=strio)
        strio.seek(0)
        with Image(file=strio) as png:
            assert png.format == 'PNG'
        with raises(ValueError):
            img.format = 'HONG'
        with raises(TypeError):
            img.format = 123


def test_fuzz():
    with Image(filename='rose:') as img:
        assert img.fuzz == 0.0
        img.fuzz = img.quantum_range
        assert img.fuzz == img.quantum_range


def test_gravity_set():
    with Image(width=144, height=192, background=Color('#1e50a2')) as img:
        img.gravity = 'center'
        assert img.gravity == 'center'


def test_histogram(fx_asset):
    with Image(filename=str(fx_asset.join('trim-color-test.png'))) as a:
        h = a.histogram
        assert len(h) == 2
        assert frozenset(h) == frozenset([
            Color('srgb(0,255,0'),
            Color('srgb(0,0,255')
        ])
        assert dict(h) == {
            Color('srgb(0,255,0'): 5000,
            Color('srgb(0,0,255'): 5000,
        }
        assert Color('white') not in h
        assert Color('srgb(0,255,0)') in h
        assert Color('srgb(0,0,255)') in h
        assert h[Color('srgb(0,255,0)')] == 5000
        assert h[Color('srgb(0,0,255)')] == 5000


def test_interlace_scheme_get(fx_asset):
    with Image(filename='rose:') as img:
        expected = 'no'
        assert img.interlace_scheme == expected


def test_interlace_scheme_set(fx_asset):
    with Image(filename='rose:') as img:
        expected = 'plane'
        img.interlace_scheme = expected
        assert img.interlace_scheme == expected


def test_interpolate_method_get(fx_asset):
    with Image(filename='rose:') as img:
        expected = 'undefined'
        assert img.interpolate_method == expected


def test_interpolate_method_set(fx_asset):
    with Image(filename='rose:') as img:
        expected = 'spline'
        img.interpolate_method = expected
        assert img.interpolate_method == expected


def test_kurtosis():
    with Image(filename='rose:') as img:
        kurtosis = img.kurtosis
        assert isinstance(kurtosis, numbers.Real)
        assert kurtosis != 0.0


def test_length_of_bytes():
    with Image(filename='rose:') as img:
        assert img.length_of_bytes > 0
        img.resample(300, 300)
        assert img.length_of_bytes == 0


def test_loop(fx_asset):
    with Image(filename=str(fx_asset.join('nocomments.gif'))) as img:
        assert img.loop == 0
        img.loop = 1
        assert img.loop == 1


def test_matte_color(fx_asset):
    with Image(filename='rose:') as img:
        with Color('navy') as color:
            img.matte_color = color
            assert img.matte_color == color
            with raises(TypeError):
                img.matte_color = False
        img.matte_color = 'orange'
        assert img.matte_color == Color('orange')


def test_mean():
    with Image(filename='rose:') as img:
        mean = img.mean
        assert isinstance(mean, numbers.Real)
        assert mean != 0.0


def test_metadata(fx_asset):
    """Test metadata api"""
    with Image(filename=str(fx_asset.join('beach.jpg'))) as img:
        assert 52 <= len(img.metadata) <= 55
        for key in img.metadata:
            assert isinstance(key, string_type)
        assert 'exif:ApertureValue' in img.metadata
        assert 'exif:UnknownValue' not in img.metadata
        assert img.metadata['exif:ApertureValue'] == '192/32'
        assert img.metadata.get('exif:UnknownValue', "IDK") == "IDK"


def test_mimetype(fx_asset):
    """Gets mimetypes of the image."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        assert img.mimetype in ('image/jpeg', 'image/x-jpeg')
    with Image(filename=str(fx_asset.join('croptest.png'))) as img:
        assert img.mimetype in ('image/png', 'image/x-png')


def test_minima_maxima():
    with Image(filename='rose:') as img:
        minq = img.minima
        maxq = img.maxima
        assert minq < maxq


def test_orientation_get(fx_asset):
    with Image(filename=str(fx_asset.join('sasha.jpg'))) as img:
        assert img.orientation == 'undefined'

    with Image(filename=str(fx_asset.join('beach.jpg'))) as img:
        assert img.orientation == 'top_left'


def test_orientation_set(fx_asset):
    with Image(filename=str(fx_asset.join('beach.jpg'))) as img:
        img.orientation = 'bottom_right'
        assert img.orientation == 'bottom_right'


def test_page_basic(fx_asset):
    with Image(filename=str(fx_asset.join('watermark.png'))) as img1:
        assert img1.page == (640, 480, 0, 0)
        assert img1.page_width == 640
        assert img1.page_height == 480
        assert img1.page_x == 0
        assert img1.page_y == 0
        with raises(TypeError):
            img1.page = 640


def test_page_offset(fx_asset):
    with Image(filename=str(fx_asset.join('watermark-offset.png'))) as img1:
        assert img1.page == (640, 480, 12, 13)
        assert img1.page_width == 640
        assert img1.page_height == 480
        assert img1.page_x == 12
        assert img1.page_y == 13


def test_page_setter(fx_asset):
    with Image(filename=str(fx_asset.join('watermark.png'))) as img1:
        assert img1.page == (640, 480, 0, 0)
        img1.page = (640, 480, 0, 0)
        assert img1.page == (640, 480, 0, 0)
        img1.page = (640, 480, 12, 13)
        assert img1.page == (640, 480, 12, 13)
        img1.page = (640, 480, -12, 13)
        assert img1.page == (640, 480, -12, 13)
        img1.page = (640, 480, 12, -13)
        assert img1.page == (640, 480, 12, -13)
        img1.page = (6400, 4800, 2, 3)
        assert img1.page == (6400, 4800, 2, 3)


def test_page_setter_items(fx_asset):
    with Image(filename=str(fx_asset.join('watermark.png'))) as img1:
        assert img1.page == (640, 480, 0, 0)
        img1.page_width = 6400
        assert img1.page == (6400, 480, 0, 0)
        img1.page_height = 4800
        assert img1.page == (6400, 4800, 0, 0)
        img1.page_x = 12
        assert img1.page == (6400, 4800, 12, 0)
        img1.page_y = 13
        assert img1.page == (6400, 4800, 12, 13)
        img1.page_x = -12
        assert img1.page == (6400, 4800, -12, 13)
        img1.page_y = -13
        assert img1.page == (6400, 4800, -12, -13)


def test_primary_points(fx_asset):
    with Image(filename='rose:') as img:
        blue = [d/2 for d in img.blue_primary]
        img.blue_primary = blue
        assert blue == list(img.blue_primary)
        green = [d/2 for d in img.green_primary]
        img.green_primary = green
        assert green == list(img.green_primary)
        red = [d/2 for d in img.red_primary]
        img.red_primary = red
        assert red == list(img.red_primary)
        white = [d/2 for d in img.white_point]
        img.white_point = white
        assert white == list(img.white_point)
        with raises(TypeError):
            img.blue_primary = 0xDEADBEEF
        with raises(TypeError):
            img.green_primary = 0xDEADBEEF
        with raises(TypeError):
            img.red_primary = 0xDEADBEEF
        with raises(TypeError):
            img.white_point = 0xDEADBEEF


def test_profiles(fx_asset):
    with Image(filename=str(fx_asset.join('beach.jpg'))) as img:
        assert len(img.profiles) == 1
        assert 'exif' in [d for d in img.profiles]
        exif_data = img.profiles['exif']
        assert exif_data is not None
        del img.profiles['exif']
        assert img.profiles['exif'] is None
        img.profiles['exif'] = exif_data
        assert img.profiles['exif'] == exif_data
        with raises(TypeError):
            img.profiles[0xDEADBEEF]
        with raises(TypeError):
            del img.profiles[0xDEADBEEF]
        with raises(TypeError):
            img.profiles[0xDEADBEEF] = 0xDEADBEEF
        with raises(TypeError):
            img.profiles['exif'] = 0xDEADBEEF


def test_rendering_intent(fx_asset):
    with Image(filename=str(fx_asset.join('trimtest.png'))) as img:
        assert img.rendering_intent == 'perceptual'
        img.rendering_intent = 'relative'
        assert img.rendering_intent == 'relative'


def test_resolution_get(fx_asset):
    """Gets image resolution."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        assert img.resolution == (72, 72)


def test_resolution_set_01(fx_asset):
    """Sets image resolution."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        img.resolution = (100, 100)
        assert img.resolution == (100, 100)


def test_resolution_set_02(fx_asset):
    """Sets image resolution with integer as parameter."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        img.resolution = 100
        assert img.resolution == (100, 100)


@mark.pdf
def test_resolution_set_03(fx_asset):
    """Sets image resolution on constructor"""
    try:
        with Image(filename=str(fx_asset.join('sample.pdf')),
                   resolution=(100, 100)) as img:
            assert img.resolution == (100, 100)
    except DelegateError:
        warnings.warn('PDF delegate could not be found.')


@mark.pdf
def test_resolution_set_04(fx_asset):
    """Sets image resolution on constructor with integer as parameter."""
    try:
        with Image(filename=str(fx_asset.join('sample.pdf')),
                   resolution=100) as img:
            assert img.resolution == (100, 100)
    except DelegateError:
        warnings.warn('PDF delegate could not be found.')


def test_scene():
    with Image(filename='rose:') as img:
        img.scene = 4
        assert img.scene == 4


def test_signature(fx_asset):
    """Gets the image signature."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        with fx_asset.join('mona-lisa.jpg').open('rb') as f:
            with Image(file=f) as same:
                assert img.signature == same.signature
        with img.convert('png') as same:
            assert img.signature == same.signature
        with Image(filename=str(fx_asset.join('beach.jpg'))) as diff:
            assert img.signature != diff.signature


def test_size(fx_asset):
    """Gets the image size."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        assert img.size == (402, 599)
        assert img.width == 402
        assert img.height == 599
        assert len(img) == 599


def test_skewness():
    with Image(filename='rose:') as img:
        skewness = img.skewness
        assert isinstance(skewness, numbers.Real)
        assert skewness != 0.0


def test_standard_deviation():
    with Image(filename='rose:') as img:
        standard_deviation = img.standard_deviation
        assert isinstance(standard_deviation, numbers.Real)
        assert standard_deviation != 0.0


def test_stroke_color_user_error():
    with Image(filename='rose:') as img:
        img.stroke_color = 'green'
        img.stroke_color = None
        assert img.stroke_color is None
        with raises(TypeError):
            img.stroke_color = 0xDEADBEEF


def test_type_get(fx_asset):
    """Gets the image type."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        assert img.type == "truecolor"
        img.alpha_channel = True
        if MAGICK_VERSION_NUMBER < 0x700:
            expected = "truecolormatte"
        else:
            expected = "truecoloralpha"
        assert img.type == expected


def test_type_set(fx_asset):
    """Sets the image type."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        img.type = "grayscale"
        assert img.type == "grayscale"


def test_ticks_per_second(fx_asset):
    with Image(filename=str(fx_asset.join('nocomments.gif'))) as img:
        assert img.ticks_per_second == 100
        img.ticks_per_second = 10
        assert img.ticks_per_second == 10


def test_units_get(fx_asset):
    """Gets the image resolution units."""
    with Image(filename=str(fx_asset.join('beach.jpg'))) as img:
        assert img.units == "pixelsperinch"
    with Image(filename=str(fx_asset.join('sasha.jpg'))) as img:
        assert img.units == "undefined"


def test_units_set(fx_asset):
    """Sets the image resolution units."""
    with Image(filename=str(fx_asset.join('watermark.png'))) as img:
        img.units = "pixelspercentimeter"
        assert img.units == "pixelspercentimeter"


def test_virtual_pixel_get(fx_asset):
    """Gets image virtual pixel"""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        assert img.virtual_pixel == "undefined"


def test_virtual_pixel_set(fx_asset):
    """Sets image virtual pixel"""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        img.virtual_pixel = "tile"
        assert img.virtual_pixel == "tile"
        with raises(ValueError):
            img.virtual_pixel = "nothing"



import datetime
import numbers
import re
from py.test import mark, raises

from wand.version import (MAGICK_VERSION, MAGICK_VERSION_INFO,
                          MAGICK_VERSION_NUMBER, MAGICK_RELEASE_DATE,
                          MAGICK_RELEASE_DATE_STRING, QUANTUM_DEPTH,
                          configure_options, fonts, formats)


def test_version():
    """Test version strings."""
    match = re.match(r'^ImageMagick\s+\d+\.\d+\.\d+(?:-\d+)?', MAGICK_VERSION)
    assert match
    assert isinstance(MAGICK_VERSION_INFO, tuple)
    assert (len(MAGICK_VERSION_INFO) ==
            match.group(0).count('.') + match.group(0).count('-') + 1)
    assert all(isinstance(v, int) for v in MAGICK_VERSION_INFO)
    assert isinstance(MAGICK_VERSION_NUMBER, numbers.Integral)
    assert isinstance(MAGICK_RELEASE_DATE, datetime.date)
    date_strings = (MAGICK_RELEASE_DATE.strftime('%Y-%m-%d'),
                    MAGICK_RELEASE_DATE.strftime('%Y%m%d'))
    assert MAGICK_RELEASE_DATE_STRING in date_strings


def test_quantum_depth():
    """QUANTUM_DEPTH must be one of 8, 16, 32, or 64."""
    assert QUANTUM_DEPTH in (8, 16, 32, 64)


def test_configure_options():
    assert 'RELEASE_DATE' in configure_options('RELEASE_DATE')
    with raises(TypeError):
        configure_options(0xDEADBEEF)


def test_fonts():
    font_list = fonts()
    if not font_list:
        mark.skip('Fonts not configured on system')
    else:
        first_font = font_list[0]
        first_font_part = first_font[1:-1]
        assert first_font in fonts('*{0}*'.format(first_font_part))
    with raises(TypeError):
        fonts(0xDEADBEEF)


def test_formats():
    xc = 'XC'
    assert formats(xc) == [xc]
    with raises(TypeError):
        formats(0xDEADBEEF)



""":mod:`wand.cdefs.drawing_wand` --- Drawing-Wand definitions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. versionadded:: 0.5.0
"""
from ctypes import (POINTER, c_void_p, c_char_p, c_double, c_int, c_uint,
                    c_size_t, c_ubyte, c_ulong)
from wand.cdefs.wandtypes import c_magick_char_p, c_ssize_t
from wand.cdefs.structures import PointInfo

__all__ = ('load',)


def load(lib, IM_VERSION):
    """Define Drawing Wand methods. The ImageMagick version is given as a
    second argument for comparison. This will quick to determine which methods
    are available from the library, and can be implemented as::

        if IM_VERSION < 0x700:
            # ... do ImageMagick-6 methods ...
        else
            # ... do ImageMagick-7 methods ...

    .. seealso::

        #include "wand/drawing-wand.h"
        // Or
        #include "MagickWand/drawing-wand.h"

    :param lib: the loaded ``MagickWand`` library
    :type lib: :class:`ctypes.CDLL`
    :param IM_VERSION: the ImageMagick version number (i.e. 0x0689)
    :type IM_VERSION: :class:`numbers.Integral`

    .. versionadded:: 0.5.0

    """
    is_im_6 = IM_VERSION < 0x700
    is_im_7 = IM_VERSION >= 0x700
    lib.NewDrawingWand.restype = c_void_p
    lib.CloneDrawingWand.argtypes = [c_void_p]
    lib.CloneDrawingWand.restype = c_void_p
    lib.DestroyDrawingWand.argtypes = [c_void_p]
    lib.DestroyDrawingWand.restype = c_void_p
    lib.IsDrawingWand.argtypes = [c_void_p]
    lib.IsDrawingWand.restype = c_int
    lib.DrawGetException.argtypes = [c_void_p, POINTER(c_int)]
    lib.DrawGetException.restype = c_magick_char_p
    lib.DrawClearException.argtypes = [c_void_p]
    lib.DrawClearException.restype = c_int
    lib.DrawAffine.argtypes = [c_void_p, c_void_p]
    if is_im_7:
        lib.DrawAlpha.argtypes = [c_void_p, c_double, c_double, c_int]
    lib.DrawComment.argtypes = [c_void_p, c_char_p]
    lib.DrawComposite.argtypes = [
        c_void_p, c_int, c_double, c_double, c_double, c_double, c_void_p
    ]
    lib.DrawComposite.restype = c_uint
    lib.DrawSetBorderColor.argtypes = [c_void_p, c_void_p]
    lib.DrawSetClipPath.argtypes = [c_void_p, c_char_p]
    lib.DrawSetClipPath.restype = c_int
    lib.DrawSetClipRule.argtypes = [c_void_p, c_uint]
    lib.DrawSetClipUnits.argtypes = [c_void_p, c_uint]
    lib.DrawSetFont.argtypes = [c_void_p, c_char_p]
    lib.DrawSetFontFamily.argtypes = [c_void_p, c_char_p]
    lib.DrawSetFontFamily.restype = c_uint
    lib.DrawSetFontResolution.argtypes = [c_void_p, c_double, c_double]
    lib.DrawSetFontResolution.restype = c_uint
    lib.DrawSetFontSize.argtypes = [c_void_p, c_double]
    lib.DrawSetFontStretch.argtypes = [c_void_p, c_int]
    lib.DrawSetFontStyle.argtypes = [c_void_p, c_int]
    lib.DrawSetFontWeight.argtypes = [c_void_p, c_size_t]
    lib.DrawSetFillColor.argtypes = [c_void_p, c_void_p]
    lib.DrawSetFillOpacity.argtypes = [c_void_p, c_double]
    lib.DrawSetFillPatternURL.argtypes = [c_void_p, c_char_p]
    lib.DrawSetFillPatternURL.restype = c_uint
    lib.DrawSetFillRule.argtypes = [c_void_p, c_uint]
    lib.DrawSetOpacity.argtypes = [c_void_p, c_double]
    lib.DrawSetStrokeAntialias.argtypes = [c_void_p, c_int]
    lib.DrawSetStrokeColor.argtypes = [c_void_p, c_void_p]
    lib.DrawSetStrokeDashArray.argtypes = [
        c_void_p, c_size_t, POINTER(c_double)
    ]
    lib.DrawSetStrokeDashOffset.argtypes = [c_void_p, c_double]
    lib.DrawSetStrokeLineCap.argtypes = [c_void_p, c_int]
    lib.DrawSetStrokeLineJoin.argtypes = [c_void_p, c_int]
    lib.DrawSetStrokeMiterLimit.argtypes = [c_void_p, c_size_t]
    lib.DrawSetStrokeOpacity.argtypes = [c_void_p, c_double]
    lib.DrawSetStrokePatternURL.argtypes = [c_void_p, c_char_p]
    lib.DrawSetStrokePatternURL.restype = c_uint
    lib.DrawSetStrokeWidth.argtypes = [c_void_p, c_double]
    lib.DrawSetTextAlignment.argtypes = [c_void_p, c_int]
    lib.DrawSetTextAntialias.argtypes = [c_void_p, c_int]
    lib.DrawSetTextDecoration.argtypes = [c_void_p, c_int]
    try:
        lib.DrawSetTextDirection.argtypes = [c_void_p, c_int]
    except AttributeError:
        lib.DrawSetTextDirection = None
    lib.DrawSetTextEncoding.argtypes = [c_void_p, c_char_p]
    try:
        lib.DrawSetTextInterlineSpacing.argtypes = [c_void_p, c_double]
    except AttributeError:
        lib.DrawSetTextInterlineSpacing = None
    lib.DrawSetTextInterwordSpacing.argtypes = [c_void_p, c_double]
    lib.DrawSetTextKerning.argtypes = [c_void_p, c_double]
    lib.DrawSetTextUnderColor.argtypes = [c_void_p, c_void_p]
    lib.DrawSetVectorGraphics.argtypes = [c_void_p, c_char_p]
    lib.DrawSetVectorGraphics.restype = c_int
    lib.DrawResetVectorGraphics.argtypes = [c_void_p]
    lib.DrawSetViewbox.argtypes = [
        c_void_p, c_ssize_t, c_ssize_t, c_ssize_t, c_ssize_t
    ]
    lib.DrawGetBorderColor.argtypes = [c_void_p, c_void_p]
    lib.DrawGetClipPath.argtypes = [c_void_p]
    lib.DrawGetClipPath.restype = c_magick_char_p
    lib.DrawGetClipRule.argtypes = [c_void_p]
    lib.DrawGetClipRule.restype = c_uint
    lib.DrawGetClipUnits.argtypes = [c_void_p]
    lib.DrawGetClipUnits.restype = c_uint
    lib.DrawGetFillColor.argtypes = [c_void_p, c_void_p]
    lib.DrawGetFillOpacity.argtypes = [c_void_p]
    lib.DrawGetFillOpacity.restype = c_double
    lib.DrawGetFillRule.argtypes = [c_void_p]
    lib.DrawGetFillRule.restype = c_uint
    lib.DrawGetOpacity.argtypes = [c_void_p]
    lib.DrawGetOpacity.restype = c_double
    lib.DrawGetStrokeAntialias.argtypes = [c_void_p]
    lib.DrawGetStrokeAntialias.restype = c_int
    lib.DrawGetStrokeColor.argtypes = [c_void_p, c_void_p]
    lib.DrawGetStrokeDashArray.argtypes = [c_void_p, POINTER(c_size_t)]
    lib.DrawGetStrokeDashArray.restype = POINTER(c_double)
    lib.DrawGetStrokeDashOffset.argtypes = [c_void_p]
    lib.DrawGetStrokeDashOffset.restype = c_double
    lib.DrawGetStrokeLineCap.argtypes = [c_void_p]
    lib.DrawGetStrokeLineCap.restype = c_int
    lib.DrawGetStrokeLineJoin.argtypes = [c_void_p]
    lib.DrawGetStrokeLineJoin.restype = c_int
    lib.DrawGetStrokeMiterLimit.argtypes = [c_void_p]
    lib.DrawGetStrokeMiterLimit.restype = c_size_t
    lib.DrawGetStrokeOpacity.argtypes = [c_void_p]
    lib.DrawGetStrokeOpacity.restype = c_double
    lib.DrawGetStrokeWidth.argtypes = [c_void_p]
    lib.DrawGetStrokeWidth.restype = c_double
    lib.DrawGetFont.argtypes = [c_void_p]
    lib.DrawGetFont.restype = c_magick_char_p
    lib.DrawGetFontFamily.argtypes = [c_void_p]
    lib.DrawGetFontFamily.restype = c_magick_char_p
    lib.DrawGetFontResolution.argtypes = [
        c_void_p, POINTER(c_double), POINTER(c_double)
    ]
    lib.DrawGetFontResolution.restype = c_uint
    lib.DrawGetFontSize.argtypes = [c_void_p]
    lib.DrawGetFontSize.restype = c_double
    lib.DrawGetFontStyle.argtypes = [c_void_p]
    lib.DrawGetFontStyle.restype = c_int
    lib.DrawGetFontWeight.argtypes = [c_void_p]
    lib.DrawGetFontWeight.restype = c_size_t
    lib.DrawGetFontStretch.argtypes = [c_void_p]
    lib.DrawGetFontStretch.restype = c_int
    lib.DrawGetTextAlignment.argtypes = [c_void_p]
    lib.DrawGetTextAlignment.restype = c_int
    lib.DrawGetTextAntialias.argtypes = [c_void_p]
    lib.DrawGetTextAntialias.restype = c_int
    lib.DrawGetTextDecoration.argtypes = [c_void_p]
    lib.DrawGetTextDecoration.restype = c_int
    try:
        lib.DrawGetTextDirection.argtypes = [c_void_p]
        lib.DrawGetTextDirection.restype = c_int
    except AttributeError:
        lib.DrawGetTextDirection = None
    lib.DrawGetTextEncoding.argtypes = [c_void_p]
    lib.DrawGetTextEncoding.restype = c_magick_char_p
    try:
        lib.DrawGetTextInterlineSpacing.argtypes = [c_void_p]
        lib.DrawGetTextInterlineSpacing.restype = c_double
    except AttributeError:
        lib.DrawGetTextInterlineSpacing = None
    lib.DrawGetTextInterwordSpacing.argtypes = [c_void_p]
    lib.DrawGetTextInterwordSpacing.restype = c_double
    lib.DrawGetTextKerning.argtypes = [c_void_p]
    lib.DrawGetTextKerning.restype = c_double
    lib.DrawGetTextUnderColor.argtypes = [c_void_p, c_void_p]
    lib.DrawGetVectorGraphics.argtypes = [c_void_p]
    lib.DrawGetVectorGraphics.restype = c_magick_char_p
    lib.DrawSetGravity.argtypes = [c_void_p, c_int]
    lib.DrawGetGravity.argtypes = [c_void_p]
    lib.DrawGetGravity.restype = c_int
    lib.ClearDrawingWand.argtypes = [c_void_p]
    lib.DrawAnnotation.argtypes = [
        c_void_p, c_double, c_double, POINTER(c_ubyte)
    ]
    lib.DrawArc.argtypes = [
        c_void_p, c_double, c_double, c_double, c_double, c_double, c_double
    ]
    lib.DrawBezier.argtypes = [c_void_p, c_ulong, POINTER(PointInfo)]
    lib.DrawCircle.argtypes = [
        c_void_p, c_double, c_double, c_double, c_double
    ]
    lib.DrawColor.argtypes = [c_void_p, c_double, c_double, c_uint]
    lib.DrawEllipse.argtypes = [
        c_void_p, c_double, c_double, c_double, c_double, c_double, c_double
    ]
    lib.DrawLine.argtypes = [c_void_p, c_double, c_double, c_double, c_double]
    if is_im_6:
        lib.DrawMatte.argtypes = [c_void_p, c_double, c_double, c_int]
    else:
        lib.DrawMatte = None
    lib.DrawPathClose.argtypes = [c_void_p]
    lib.DrawPathCurveToAbsolute.argtypes = [
        c_void_p, c_double, c_double, c_double, c_double, c_double, c_double
    ]
    lib.DrawPathCurveToRelative.argtypes = [
        c_void_p, c_double, c_double, c_double, c_double, c_double, c_double
    ]
    lib.DrawPathCurveToQuadraticBezierAbsolute.argtypes = [
        c_void_p, c_double, c_double, c_double, c_double
    ]
    lib.DrawPathCurveToQuadraticBezierRelative.argtypes = [
        c_void_p, c_double, c_double, c_double, c_double
    ]
    lib.DrawPathCurveToQuadraticBezierSmoothAbsolute.argtypes = [
        c_void_p, c_double, c_double
    ]
    lib.DrawPathCurveToQuadraticBezierSmoothRelative.argtypes = [
        c_void_p, c_double, c_double
    ]
    lib.DrawPathCurveToSmoothAbsolute.argtypes = [
        c_void_p, c_double, c_double, c_double, c_double
    ]
    lib.DrawPathCurveToSmoothRelative.argtypes = [
        c_void_p, c_double, c_double, c_double, c_double
    ]
    lib.DrawPathEllipticArcAbsolute.argtypes = [
        c_void_p, c_double, c_double, c_double, c_uint, c_uint, c_double,
        c_double
    ]
    lib.DrawPathEllipticArcRelative.argtypes = [
        c_void_p, c_double, c_double, c_double, c_uint, c_uint, c_double,
        c_double
    ]
    lib.DrawPathFinish.argtypes = [c_void_p]
    lib.DrawPathLineToAbsolute.argtypes = [c_void_p, c_double, c_double]
    lib.DrawPathLineToRelative.argtypes = [c_void_p, c_double, c_double]
    lib.DrawPathLineToHorizontalAbsolute.argtypes = [c_void_p, c_double]
    lib.DrawPathLineToHorizontalRelative.argtypes = [c_void_p, c_double]
    lib.DrawPathLineToVerticalAbsolute.argtypes = [c_void_p, c_double]
    lib.DrawPathLineToVerticalRelative.argtypes = [c_void_p, c_double]
    lib.DrawPathMoveToAbsolute.argtypes = [c_void_p, c_double, c_double]
    lib.DrawPathMoveToRelative.argtypes = [c_void_p, c_double, c_double]
    lib.DrawPathStart.argtypes = [c_void_p]
    lib.DrawPoint.argtypes = [c_void_p, c_double, c_double]
    lib.DrawPolygon.argtypes = [c_void_p, c_ulong, POINTER(PointInfo)]
    lib.DrawPolyline.argtypes = [c_void_p, c_ulong, POINTER(PointInfo)]
    lib.DrawRotate.argtypes = [c_void_p, c_double]
    lib.DrawRectangle.argtypes = [
        c_void_p, c_double, c_double, c_double, c_double
    ]
    lib.DrawRoundRectangle.argtypes = [
        c_void_p, c_double, c_double, c_double, c_double, c_double, c_double
    ]
    lib.DrawScale.argtypes = [c_void_p, c_double, c_double]
    lib.DrawSkewX.argtypes = [c_void_p, c_double]
    lib.DrawSkewY.argtypes = [c_void_p, c_double]
    lib.DrawTranslate.argtypes = [c_void_p, c_double, c_double]
    lib.PushDrawingWand.argtypes = [c_void_p]
    lib.PushDrawingWand.restype = c_uint
    lib.DrawPushClipPath.argtypes = [c_void_p, c_char_p]
    lib.DrawPushDefs.argtypes = [c_void_p]
    lib.DrawPushPattern.argtypes = [
        c_void_p, c_char_p, c_double, c_double, c_double, c_double
    ]
    lib.DrawPushClipPath.restype = c_uint
    lib.PopDrawingWand.argtypes = [c_void_p]
    lib.PopDrawingWand.restype = c_uint
    lib.DrawPopClipPath.argtypes = [c_void_p]
    lib.DrawPopDefs.argtypes = [c_void_p]
    lib.DrawPopPattern.argtypes = [c_void_p]



# -*- coding: utf-8 -*-
#
# These tests cover the basic I/O & pythonic interfaces of the Image class.
#
import codecs
import io
import os
import os.path
import shutil
import struct
import sys
import tempfile

from pytest import mark, raises

from wand.image import ClosedImageError, Image
from wand.color import Color
from wand.compat import PY3, text, text_type

try:
    filesystem_encoding = sys.getfilesystemencoding()
except RuntimeError:
    unicode_filesystem_encoding = False
else:
    try:
        codec_info = codecs.lookup(filesystem_encoding)
    except LookupError:
        unicode_filesystem_encoding = False
    else:
        unicode_filesystem_encoding = codec_info.name in (
            'utf-8', 'utf-16', 'utf-16-be', 'utf-16-le',
            'utf-32', 'utf-32-be', 'utf-32-le',
            'mbcs'  # for Windows
        )

try:
    import numpy as np
except ImportError:
    np = None


def test_empty_image():
    with Image() as img:
        assert img.size == (0, 0)
        assert repr(img) == '<wand.image.Image: (empty)>'


def test_image_invalid_params():
    with raises(TypeError):
        Image(image=Image(), width=100, height=100)
    with raises(TypeError):
        Image(image=Image(), blob=b"blob")
    with raises(TypeError):
        Image(image=b"blob")


def test_blank_image():
    gray = Color('#ccc')
    transparent = Color('transparent')
    with raises(ValueError):
        Image(width=0, height=0)
    with Image(width=20, height=10) as img:
        assert img[10, 5] == transparent
    with Image(width=20, height=10, background=gray) as img:
        assert img.size == (20, 10)
        assert img[10, 5] == gray
    with Image(width=20, height=10, background='#ccc') as img:
        assert img.size == (20, 10)
        assert img[10, 5] == gray


def test_raw_image(fx_asset):
    b = b"".join([struct.pack("BBB", i, j, 0)
                  for i in range(256) for j in range(256)])
    with raises(ValueError):
        Image(blob=b, depth=6)
    with raises(ValueError):
        Image(blob=b, depth=8, width=0, height=0, format="RGB")
    with raises(TypeError):
        Image(blob=b, depth=8, width=256, height=256, format=1)
    with Image(blob=b, depth=8, width=256, height=256, format="RGB") as img:
        assert img.size == (256, 256)
        assert img[0, 0] == Color('#000000')
        assert img[255, 255] == Color('#ffff00')
        assert img[64, 128] == Color('#804000')
    with Image(filename=str(fx_asset.join('blob.rgb')),
               depth=8, width=256, height=256, format="RGB") as img:
        assert img.size == (256, 256)
        assert img[0, 0] == Color('#000000')
        assert img[255, 255] == Color('#ffff00')
        assert img[64, 128] == Color('#804000')


def test_clear_image(fx_asset):
    with Image() as img:
        img.read(filename=str(fx_asset.join('mona-lisa.jpg')))
        assert img.size == (402, 599)
        img.clear()
        assert img.size == (0, 0)
        img.read(filename=str(fx_asset.join('beach.jpg')))
        assert img.size == (800, 600)


def test_read_from_filename(fx_asset):
    with Image() as img:
        img.read(filename=str(fx_asset.join('mona-lisa.jpg')))
        assert img.width == 402
        img.clear()
        with fx_asset.join('mona-lisa.jpg').open('rb') as f:
            img.read(file=f)
            assert img.width == 402
            img.clear()
        blob = fx_asset.join('mona-lisa.jpg').read('rb')
        img.read(blob=blob)
        assert img.width == 402


@mark.skipif(not unicode_filesystem_encoding,
             reason='Unicode filesystem encoding needed')
def test_read_from_unicode_filename(fx_asset, tmpdir):
    """https://github.com/emcconville/wand/issues/122"""
    filename = '모나리자.jpg'
    if not PY3:
        filename = filename.decode('utf-8')
    path = os.path.join(text_type(tmpdir), filename)  # workaround py.path bug
    shutil.copyfile(str(fx_asset.join('mona-lisa.jpg')), path)
    with Image() as img:
        img.read(filename=text(path))
        assert img.width == 402


@mark.pdf
def test_read_with_colorspace(fx_asset):
    fpath = str(fx_asset.join('cmyka.pdf'))
    with Image(filename=fpath,
               resolution=100,
               colorspace='srgb',
               units='pixelspercentimeter') as img:
        assert img.colorspace == 'srgb'
        assert img.units == 'pixelspercentimeter'


def test_new_from_file(fx_asset):
    """Opens an image from the file object."""
    with fx_asset.join('mona-lisa.jpg').open('rb') as f:
        with Image(file=f) as img:
            assert img.width == 402
    with raises(ClosedImageError):
        img.wand
    strio = io.BytesIO(fx_asset.join('mona-lisa.jpg').read('rb'))
    with Image(file=strio) as img:
        assert img.width == 402
    strio.close()
    with raises(ClosedImageError):
        img.wand
    with raises(TypeError):
        Image(file='not file object')


def test_new_from_filename(fx_asset):
    """Opens an image through its filename."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        assert img.width == 402
    with raises(ClosedImageError):
        img.wand
    with raises(IOError):
        Image(filename=str(fx_asset.join('not-exists.jpg')))


@mark.skipif(not unicode_filesystem_encoding,
             reason='Unicode filesystem encoding needed')
def test_new_from_unicode_filename(fx_asset, tmpdir):
    """https://github.com/emcconville/wand/issues/122"""
    filename = '모나리자.jpg'
    if not PY3:
        filename = filename.decode('utf-8')
    path = os.path.join(text_type(tmpdir), filename)  # workaround py.path bug
    shutil.copyfile(str(fx_asset.join('mona-lisa.jpg')), path)
    with Image(filename=text(path)) as img:
        assert img.width == 402


def test_new_from_blob(fx_asset):
    """Opens an image from blob."""
    blob = fx_asset.join('mona-lisa.jpg').read('rb')
    with Image(blob=blob) as img:
        assert img.width == 402
    with raises(ClosedImageError):
        img.wand


def test_new_with_format(fx_asset):
    blob = fx_asset.join('google.ico').read('rb')
    with raises(Exception):
        Image(blob=blob)
    with Image(blob=blob, format='ico') as img:
        assert img.size == (16, 16)


def test_new_from_pseudo(fx_asset):
    with Image() as img:
        img.pseudo(10, 10, 'xc:none')
        assert img.size == (10, 10)


def test_clone(fx_asset):
    """Clones the existing image."""
    funcs = (lambda img: Image(image=img),
             lambda img: img.clone())
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        for func in funcs:
            with func(img) as cloned:
                assert img.wand is not cloned.wand
                assert img.size == cloned.size
            with raises(ClosedImageError):
                cloned.wand
    with raises(ClosedImageError):
        img.wand


def test_ping_from_filename(fx_asset):
    file_path = str(fx_asset.join('mona-lisa.jpg'))
    with Image.ping(filename=file_path) as img:
        assert img.size == (402, 599)


def test_ping_from_blob(fx_asset):
    blob = fx_asset.join('mona-lisa.jpg').read('rb')
    with Image.ping(blob=blob) as img:
        assert img.size == (402, 599)


def test_ping_from_file(fx_asset):
    with fx_asset.join('mona-lisa.jpg').open('rb') as fd:
        with Image.ping(file=fd) as img:
            assert img.size == (402, 599)


def test_save_to_filename(fx_asset):
    """Saves an image to the filename."""
    savefile = os.path.join(tempfile.mkdtemp(), 'savetest.jpg')
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as orig:
        orig.save(filename=savefile)
        with raises(IOError):
            orig.save(filename=os.path.join(savefile, 'invalid.jpg'))
        with raises(TypeError):
            orig.save(filename=1234)
    assert os.path.isfile(savefile)
    with Image(filename=savefile) as saved:
        assert saved.size == (402, 599)
    os.remove(savefile)


@mark.skipif(not unicode_filesystem_encoding,
             reason='Unicode filesystem encoding needed')
def test_save_to_unicode_filename(fx_asset, tmpdir):
    filename = '모나리자.jpg'
    if not PY3:
        filename = filename.decode('utf-8')
    path = os.path.join(text_type(tmpdir), filename)  # workaround py.path bug
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as orig:
        orig.save(filename=path)
    with Image(filename=path) as img:
        assert img.width == 402


def test_save_to_file(fx_asset):
    """Saves an image to the Python file object."""
    buffer = io.BytesIO()
    with tempfile.TemporaryFile() as savefile:
        with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as orig:
            orig.save(file=savefile)
            orig.save(file=buffer)
            with raises(TypeError):
                orig.save(file='filename')
            with raises(TypeError):
                orig.save(file=1234)
        savefile.seek(0)
        with Image(file=savefile) as saved:
            assert saved.size == (402, 599)
        buffer.seek(0)
        with Image(file=buffer) as saved:
            assert saved.size == (402, 599)
    buffer.close()


def test_save_full_animated_gif_to_file(fx_asset):
    """Save all frames of an animated to a Python file object."""
    temp_filename = os.path.join(tempfile.mkdtemp(), 'savetest.gif')
    orig_filename = str(fx_asset.join('nocomments.gif'))
    with open(temp_filename, 'w+b') as fp:
        with Image(filename=orig_filename) as orig:
            orig.save(file=fp)
    assert os.path.isfile(temp_filename)
    with Image(filename=orig_filename) as orig:
        with Image(filename=temp_filename) as temp:
            assert len(orig.sequence) == len(temp.sequence)
    os.remove(temp_filename)


def test_save_error(fx_asset):
    filename = os.path.join(tempfile.mkdtemp(), 'savetest.jpg')
    fileobj = io.BytesIO()
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as orig:
        with raises(TypeError):
            orig.save()
        with raises(TypeError):
            orig.save(filename=filename, file=fileobj)


def test_make_blob(fx_asset):
    """Makes a blob string."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        with Image(blob=img.make_blob('png')) as img2:
            assert img2.size == (402, 599)
            assert img2.format == 'PNG'
        assert img.format == 'JPEG'
        with raises(TypeError):
            img.make_blob(123)
    svg = b'''
    <svg width="100px" height="100px">
        <circle cx="100" cy="50" r="40" stroke="black"
         stroke-width="2" fill="red" />
    </svg>
    '''
    with Image(blob=svg, format='svg') as img:
        assert img.size == (100, 100)
        assert img.format in ('SVG', 'MVG')
        img.format = 'PNG'
        assert img.size == (100, 100)
        assert img.format == 'PNG'
        png = img.make_blob()
    with Image(blob=png, format='png') as img:
        assert img.size == (100, 100)
        assert img.format == 'PNG'


def test_convert(fx_asset):
    """Converts the image format."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        with img.convert('png') as converted:
            assert converted.format == 'PNG'
            strio = io.BytesIO()
            converted.save(file=strio)
            strio.seek(0)
            with Image(file=strio) as png:
                assert png.format == 'PNG'
        with raises(ValueError):
            img.convert('HONG')
        with raises(TypeError):
            img.convert(123)


@mark.slow
def test_iterate(fx_asset):
    """Uses iterator."""
    with Color('#000') as black:
        with Color('transparent') as transparent:
            with Image(filename=str(fx_asset.join('croptest.png'))) as img:
                for i, row in enumerate(img):
                    assert len(row) == 300
                    if i % 30:
                        continue  # avoid slowness
                    if 100 <= i < 200:
                        for x, color in enumerate(row):
                            if x % 30:
                                continue  # avoid slowness
                            if 100 <= x < 200:
                                assert color == black
                            else:
                                assert color == transparent
                    else:
                        for color in row:
                            assert color == transparent
                assert i == 299


def test_slice_clone(fx_asset):
    """Clones using slicing."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        with img[:, :] as cloned:
            assert img.size == cloned.size


def test_slice_invalid_types(fx_asset):
    """Slicing with invalid types should throw exceptions."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        with raises(TypeError):
            img['12']
        with raises(TypeError):
            img[1.23]
        with raises(ValueError):
            img[()]
        with raises(ValueError):
            img[:, :, :]
        with raises(ValueError):
            img[::2, :]
        with raises(IndexError):
            img[1:1, :]
        with raises(IndexError):
            img[:, 2:2]
        with raises(TypeError):
            img[100.0:, 100.0]
        with raises(TypeError):
            img['100':, '100']
        with raises(IndexError):
            img[500:, 900]
        with raises(TypeError):
            img['1', 0]
        with raises(TypeError):
            img[1, '0']
    with Image(filename=str(fx_asset.join('croptest.png'))) as img:
        with raises(IndexError):
            img[300, 300]
        with raises(IndexError):
            img[-301, -301]


def test_index_pixel(fx_asset):
    """Gets a pixel."""
    with Image(filename=str(fx_asset.join('croptest.png'))) as img:
        assert img[0, 0] == Color('transparent')
        assert img[99, 99] == Color('transparent')
        assert img[100, 100] == Color('black')
        assert img[150, 150] == Color('black')
        assert img[-200, -200] == Color('black')
        assert img[-201, -201] == Color('transparent')


def test_index_pixel_set(fx_asset):
    with Image(filename='rose:') as img:
        with Color('black') as dot:
            img[0, 0] = dot
            assert img[0, 0] == dot
            img[0, 0] = 'black'
            assert img[0, 0] == dot
        img.colorspace = 'gray'
        with Color('gray50') as dot:
            img[0, 0] = dot
            assert img[0, 0] == dot
        img.colorspace = 'cmyk'
        with Color('cmyk(255, 0, 0, 0') as dot:
            img[0, 0] = dot
            assert img[0, 0] == dot
        with raises(TypeError):
            img[0, 0] = 1
        with raises(TypeError):
            img[0xDEADBEEF] = Color('black')
        with raises(ValueError):
            img[1, 2, 3] = Color('black')
        with raises(TypeError):
            img[0.5, "d"] = Color('black')


def test_index_row(fx_asset):
    """Gets a row."""
    with Color('transparent') as transparent:
        with Color('black') as black:
            with Image(filename=str(fx_asset.join('croptest.png'))) as img:
                for c in img[0]:
                    assert c == transparent
                for c in img[99]:
                    assert c == transparent
                for i, c in enumerate(img[100]):
                    if 100 <= i < 200:
                        assert c == black
                    else:
                        assert c == transparent
                for i, c in enumerate(img[150]):
                    if 100 <= i < 200:
                        assert c == black
                    else:
                        assert c == transparent
                for i, c in enumerate(img[-200]):
                    if 100 <= i < 200:
                        assert c == black
                    else:
                        assert c == transparent
                for c in img[-201]:
                    assert c == transparent


def test_slice_crop(fx_asset):
    """Crops using slicing."""
    with Image(filename=str(fx_asset.join('croptest.png'))) as img:
        with img[100:200, 100:200] as cropped:
            assert cropped.size == (100, 100)
            with Color('#000') as black:
                for row in cropped:
                    for col in row:
                        assert col == black
        with img[150:, :150] as cropped:
            assert cropped.size == (150, 150)
        with img[-200:-100, -200:-100] as cropped:
            assert cropped.size == (100, 100)
        with img[100:200] as cropped:
            assert cropped.size == (300, 100)
        assert img.size == (300, 300)
        with raises(IndexError):
            img[:500, :500]
        with raises(IndexError):
            img[290:310, 290:310]


def test_equal(fx_asset):
    """Equals (``==``) and not equals (``!=``) operators."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as a:
        with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as a2:
            assert a == a2
            assert not (a != a2)
        with Image(filename=str(fx_asset.join('sasha.jpg'))) as b:
            assert a != b
            assert not (a == b)
        with a.convert('png') as a3:
            assert a == a3
            assert not (a != a3)


def test_object_hash(fx_asset):
    """Gets :func:`hash()` of the image."""
    with Image(filename=str(fx_asset.join('mona-lisa.jpg'))) as img:
        a = hash(img)
        img.format = 'png'
        b = hash(img)
        assert a == b


def test_issue_150(fx_asset, tmpdir):
    """Should not be terminated with segmentation fault.

    https://github.com/emcconville/wand/issues/150

    """
    with Image(filename=str(fx_asset.join('tiger_hd-1920x1080.jpg'))) as img:
        img.format = 'pjpeg'
        with open(str(tmpdir.join('out.jpg')), 'wb') as f:
            img.save(file=f)


@mark.skipif(np is None, reason='Numpy not available.')
def test_from_array():
    # From float values.
    rand = np.random.rand(10, 10, 4)
    # We should have a 10x10 image with RGBA data created.
    with Image.from_array(rand) as img:
        assert img.size == (10, 10)
        assert img.alpha_channel
    # From char values
    red8 = np.zeros([10, 10, 3], dtype=np.uint8)
    red8[:, :] = [0xFF, 0x00, 0x00]
    # We should have a RED image.
    with Image.from_array(red8) as img:
        assert img[0, 0] == Color('#F00')
    # From short values.
    green16 = np.zeros([10, 10, 3], dtype=np.uint16)
    green16[:, :] = [0x0000, 0xFFFF, 0x0000]
    # We should have a GREEN image.
    with Image.from_array(green16) as img:
        assert img[0, 0] == Color('#00FF00')


@mark.skipif(np is None, reason='Numpy not available.')
def test_array_interface():
    with Image(filename='rose:') as img:
        img.alpha_channel = 'off'
        array = np.array(img)
        assert array.shape == (46, 70, 3)


@mark.skipif(np is None, reason='Numpy not available.')
def test_numpy_array_hairpinning():
    with Image(filename='rose:') as left:
        with Image.from_array(left) as right:
            assert left.size == right.size



""":mod:`wand.cdefs.pixel_iterator` --- Pixel-Iterator definitions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. versionadded:: 0.5.0
"""
from ctypes import POINTER, c_void_p, c_int, c_size_t
from wand.cdefs.wandtypes import c_ssize_t, c_magick_char_p

__all__ = ('load',)


def load(lib, IM_VERSION):
    """Define Pixel Iterator methods. The ImageMagick version is given as
    a second argument for comparison. This will quick to determine which
    methods are available from the library, and can be implemented as::

        if IM_VERSION < 0x700:
            # ... do ImageMagick-6 methods ...
        else
            # ... do ImageMagick-7 methods ...

    .. seealso::

        #include "wand/pixel-iterator.h"
        // Or
        #include "MagickWand/pixel-iterator.h"

    :param lib: the loaded ``MagickWand`` library
    :type lib: :class:`ctypes.CDLL`
    :param IM_VERSION: the ImageMagick version number (i.e. 0x0689)
    :type IM_VERSION: :class:`numbers.Integral`

    .. versionadded:: 0.5.0

    """
    lib.ClonePixelIterator.argtypes = [c_void_p]
    lib.ClonePixelIterator.restype = c_void_p
    lib.DestroyPixelIterator.argtypes = [c_void_p]
    lib.DestroyPixelIterator.restype = c_void_p
    lib.IsPixelIterator.argtypes = [c_void_p]
    lib.NewPixelIterator.argtypes = [c_void_p]
    lib.NewPixelIterator.restype = c_void_p
    lib.PixelClearIteratorException.argtypes = [c_void_p]
    lib.PixelGetIteratorException.argtypes = [c_void_p, POINTER(c_int)]
    lib.PixelGetIteratorException.restype = c_magick_char_p
    lib.PixelGetNextIteratorRow.argtypes = [c_void_p, POINTER(c_size_t)]
    lib.PixelGetNextIteratorRow.restype = POINTER(c_void_p)
    lib.PixelSetFirstIteratorRow.argtypes = [c_void_p]
    lib.PixelSetIteratorRow.argtypes = [c_void_p, c_ssize_t]



import itertools

from pytest import fixture, mark, raises, skip

from wand.image import Image
from wand.color import Color
from wand.api import library
from wand.drawing import Drawing
from wand.exceptions import PolicyError, WandLibraryVersionError
from wand.version import MAGICK_VERSION_NUMBER


@fixture
def fx_wand(request):
    wand = Drawing()
    request.addfinalizer(wand.destroy)
    return wand


def test_init_user_error():
    with raises(TypeError):
        with Drawing(0xDEADBEEF):
            pass


def test_is_drawing_wand(fx_wand):
    assert library.IsDrawingWand(fx_wand.resource)


def test_set_get_border_color(fx_wand):
    with Color("#0F0") as green:
        fx_wand.border_color = green
        assert green == fx_wand.border_color
    fx_wand.border_color = 'orange'
    assert fx_wand.border_color == Color('orange')
    # Assert user error
    with raises(TypeError):
        fx_wand.border_color = 0xDEADBEEF


def test_set_get_clip_path(fx_wand):
    fx_wand.clip_path = 'path_id'
    assert fx_wand.clip_path == 'path_id'
    # Assert user error
    with raises(TypeError):
        fx_wand.clip_path = 0xDEADBEEF


def test_set_get_clip_rule(fx_wand):
    fx_wand.clip_rule = 'evenodd'
    assert fx_wand.clip_rule == 'evenodd'
    with raises(TypeError):
        fx_wand.clip_rule = 0xDEADBEEF
    with raises(ValueError):
        fx_wand.clip_rule = 'not-a-rule'


def test_set_get_clip_units(fx_wand):
    fx_wand.clip_units = 'object_bounding_box'
    assert fx_wand.clip_units == 'object_bounding_box'
    with raises(TypeError):
        fx_wand.clip_units = 0xDEADBEEF
    with raises(ValueError):
        fx_wand.clip_units = 'not-a-clip_unit'


def test_set_get_font(fx_wand, fx_asset):
    fx_wand.font = str(fx_asset.join('League_Gothic.otf'))
    assert fx_wand.font == str(fx_asset.join('League_Gothic.otf'))
    with raises(TypeError):
        fx_wand.font = 0xDEADBEEF


def test_set_get_font_family(fx_wand):
    assert fx_wand.font_family is None
    fx_wand.font_family = 'sans-serif'
    assert fx_wand.font_family == 'sans-serif'
    with raises(TypeError):
        fx_wand.font_family = 0xDEADBEEF


def test_set_get_font_resolution(fx_wand):
    fx_wand.font_resolution = (78.0, 78.0)
    assert fx_wand.font_resolution == (78.0, 78.0)
    with raises(TypeError):
        fx_wand.font_resolution = 0xDEADBEEF
    with raises(ValueError):
        fx_wand.font_resolution = (78.0, 78.0, 78.0)


def test_set_get_font_size(fx_wand):
    fx_wand.font_size = 22.2
    assert fx_wand.font_size == 22.2
    with raises(TypeError):
        fx_wand.font_size = '22.2%'
    with raises(ValueError):
        fx_wand.font_size = -22.2


def test_set_get_font_stretch(fx_wand):
    fx_wand.font_stretch = 'condensed'
    assert fx_wand.font_stretch == 'condensed'
    with raises(TypeError):
        fx_wand.font_stretch = 0xDEADBEEF
    with raises(ValueError):
        fx_wand.font_stretch = 'not-a-stretch-type'


def test_set_get_font_style(fx_wand):
    fx_wand.font_style = 'italic'
    assert fx_wand.font_style == 'italic'
    with raises(TypeError):
        fx_wand.font_style = 0xDEADBEEF
    with raises(ValueError):
        fx_wand.font_style = 'not-a-style-type'


def test_set_get_font_weight(fx_wand):
    fx_wand.font_weight = 400  # Normal
    assert fx_wand.font_weight == 400
    with raises(TypeError):
        fx_wand.font_weight = '400'


def test_set_get_fill_color(fx_wand):
    with Color('#333333') as black:
        fx_wand.fill_color = black
    assert fx_wand.fill_color == Color('#333333')
    fx_wand.fill_color = 'pink'
    fx_wand.fill_color == Color('PINK')


def test_set_get_stroke_color(fx_wand):
    with Color('#333333') as black:
        fx_wand.stroke_color = black
    assert fx_wand.stroke_color == Color('#333333')
    fx_wand.stroke_color = 'skyblue'
    assert fx_wand.stroke_color == Color('SkyBlue')


def test_set_get_stroke_width(fx_wand):
    fx_wand.stroke_width = 5
    assert fx_wand.stroke_width == 5


def test_set_get_text_alignment(fx_wand):
    fx_wand.text_alignment = 'center'
    assert fx_wand.text_alignment == 'center'
    with raises(TypeError):
        fx_wand.text_alignment = 0xDEADBEEF
    with raises(ValueError):
        fx_wand.text_alignment = 'not-a-text-alignment-type'


def test_set_get_text_antialias(fx_wand):
    fx_wand.text_antialias = True
    assert fx_wand.text_antialias is True


def test_set_get_text_decoration(fx_wand):
    fx_wand.text_decoration = 'underline'
    assert fx_wand.text_decoration == 'underline'
    with raises(TypeError):
        fx_wand.text_decoration = 0xDEADBEEF
    with raises(ValueError):
        fx_wand.text_decoration = 'not-a-text-decoration-type'


def test_set_get_text_direction(fx_wand):
    try:
        fx_wand.text_direction = 'right_to_left'
        assert fx_wand.text_direction == 'right_to_left'
    except WandLibraryVersionError:
        skip("DrawGetTextDirection not supported by installed drawing library")


def test_set_get_text_encoding(fx_wand):
    fx_wand.text_encoding = 'UTF-8'
    assert fx_wand.text_encoding == 'UTF-8'
    fx_wand.text_encoding = None


def test_set_get_text_interline_spacing(fx_wand):
    fx_wand.text_interline_spacing = 10.11
    assert fx_wand.text_interline_spacing == 10.11
    with raises(TypeError):
        fx_wand.text_interline_spacing = '10.11'


def test_set_get_text_interword_spacing(fx_wand):
    fx_wand.text_interword_spacing = 5.55
    assert fx_wand.text_interword_spacing == 5.55
    with raises(TypeError):
        fx_wand.text_interline_spacing = '5.55'


def test_set_get_text_kerning(fx_wand):
    fx_wand.text_kerning = 10.22
    assert fx_wand.text_kerning == 10.22
    with raises(TypeError):
        fx_wand.text_kerning = '10.22'


def test_set_get_text_under_color(fx_wand):
    with Color('#333333') as black:
        fx_wand.text_under_color = black
    assert fx_wand.text_under_color == Color('#333333')
    fx_wand.text_under_color = '#333'  # Smoke test
    with raises(TypeError):
        fx_wand.text_under_color = 0xDEADBEEF


def test_set_get_vector_graphics(fx_wand):
    fx_wand.stroke_width = 7
    xml = fx_wand.vector_graphics
    assert xml.index("<stroke-width>7</stroke-width>") > 0
    fx_wand.vector_graphics = ('<wand><stroke-width>'
                               '8</stroke-width></wand>')
    xml = fx_wand.vector_graphics
    assert xml.index("<stroke-width>8</stroke-width>") > 0
    with raises(TypeError):
        fx_wand.vector_graphics = 0xDEADBEEF


def test_set_get_gravity(fx_wand):
    fx_wand.gravity = 'center'
    assert fx_wand.gravity == 'center'
    with raises(TypeError):
        fx_wand.gravity = 0xDEADBEEF
    with raises(ValueError):
        fx_wand.gravity = 'not-a-gravity-type'


def test_clone_drawing_wand(fx_wand):
    fx_wand.text_kerning = 10.22
    funcs = (lambda img: Drawing(drawing=fx_wand),
             lambda img: fx_wand.clone())
    for func in funcs:
        with func(fx_wand) as cloned:
            assert fx_wand.resource is not cloned.resource
            assert fx_wand.text_kerning == cloned.text_kerning


def test_clear_drawing_wand(fx_wand):
    fx_wand.text_kerning = 10.22
    assert fx_wand.text_kerning == 10.22
    fx_wand.clear()
    assert fx_wand.text_kerning == 0


def test_composite(fx_wand):
    white = Color('WHITE')
    black = Color('BLACK')
    with Image(width=50, height=50, background=white) as img:
        fx_wand.fill_color = black
        fx_wand.stroke_color = black
        fx_wand.rectangle(25, 25, 49, 49)
        fx_wand.draw(img)
        fx_wand.composite("replace", 0, 0, 25, 25, img)
        fx_wand.draw(img)
        assert img[45, 45] == img[20, 20] == black
        assert img[45, 20] == img[20, 45] == white


def test_draw_arc(fx_asset):
    white = Color('WHITE')
    red = Color('RED')
    black = Color('BLACK')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = red
            draw.stroke_color = black
            draw.arc((10, 10),   # Start
                     (40, 40),   # End
                     (-90, 90))  # Degree
            draw.draw(img)
            assert img[20, 25] == white
            assert img[30, 25] == red
            assert img[40, 25] == black


def test_draw_circle(fx_asset):
    white = Color('WHITE')
    black = Color('BLACK')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = black
            draw.circle((25, 25),  # Origin
                        (40, 40))  # Perimeter
            draw.draw(img)
            assert img[5, 5] == img[45, 45] == white
            assert img[25, 25] == black


def test_draw_comment():
    skip('TODO - Rewrite as MVG is now disabled by default on CI runner.')
    comment = 'pikachu\'s ghost'
    expected = '#pikachu\'s ghost\n'
    with Image(width=1, height=1) as img:
        with Drawing() as draw:
            draw.comment(comment)
            draw(img)
            try:
                blob = img.make_blob(format="mvg")
            except PolicyError as pe:
                skip('MVG disabled by security policies. ' + repr(pe))
            else:
                assert expected == blob


def test_draw_color():
    white = Color('WHITE')
    black = Color('BLACK')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = black
            draw.color(25, 25, 'floodfill')
            draw.draw(img)
            assert img[25, 25] == black


def test_draw_color_user_error():
    with Drawing() as draw:
        with raises(TypeError):
            draw.color('apples')
        with raises(TypeError):
            draw.color(1, 2, 4)
        with raises(ValueError):
            draw.color(1, 2, 'apples')


def test_draw_ellipse(fx_wand):
    gray, red = Color('#ccc'), Color('#f00')
    with Image(width=50, height=50, background=gray) as img:
        with Drawing() as draw:
            draw.fill_color = red
            draw.ellipse((25, 25),  # origin
                         (20, 10))  # radius
            draw.draw(img)
            assert img[25, 10] == gray
            assert img[45, 25] == red


def test_draw_line(fx_wand):
    gray = Color('#ccc')
    with Image(width=10, height=10, background=gray) as img:
        with Color('#333333') as black:
            fx_wand.fill_color = black
        fx_wand.line((5, 5), (7, 5))
        fx_wand.draw(img)
        assert img[4, 5] == Color('#ccc')
        assert img[5, 5] == Color('#333333')
        assert img[6, 5] == Color('#333333')
        assert img[7, 5] == Color('#333333')
        assert img[8, 5] == Color('#ccc')


@mark.xfail(MAGICK_VERSION_NUMBER >= 0x700,
            reason='wand.drawing.Drawing.matte removed with IM 7.',
            raises=AttributeError)
def test_draw_matte():
    white = Color('rgba(0, 255, 255, 5%)')
    transparent = Color('transparent')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_opacity = 0.0
            draw.matte(25, 25, 'floodfill')
            draw.draw(img)
            assert img[25, 25] == transparent


@mark.xfail(MAGICK_VERSION_NUMBER >= 0x700,
            reason='wand.drawing.Drawing.matte removed with IM 7.',
            raises=AttributeError)
def test_draw_matte_user_error():
    with Drawing() as draw:
        with raises(TypeError):
            draw.matte('apples')
        with raises(TypeError):
            draw.matte(1, 2, 4)
        with raises(ValueError):
            draw.matte(1, 2, 'apples')


@mark.xfail(MAGICK_VERSION_NUMBER < 0x700,
            reason='wand.drawing.Drawing.alpha was added with IM 7.',
            raises=AttributeError)
def test_draw_alpha():
    transparent = Color('transparent')
    with Image(width=50, height=50, pseudo='xc:white') as img:
        with Drawing() as draw:
            draw.fill_color = transparent
            draw.alpha(25, 25, 'floodfill')
            draw.draw(img)
        assert img[25, 25] == transparent


@mark.xfail(MAGICK_VERSION_NUMBER < 0x700,
            reason='wand.drawing.Drawing.alpha was added with IM 7.',
            raises=AttributeError)
def test_draw_alpha_user_error():
    with Drawing() as draw:
        with raises(TypeError):
            draw.alpha()
        with raises(TypeError):
            draw.alpha(1, 2, 4)
        with raises(ValueError):
            draw.alpha(1, 2, 'apples')


def test_draw_point():
    white = Color('WHITE')
    black = Color('BLACK')
    with Image(width=5, height=5, background=white) as img:
        with Drawing() as draw:
            draw.stroke_color = black
            draw.point(2, 2)
            draw.draw(img)
            assert img[2, 2] == black


def test_draw_polygon(fx_wand):
    white = Color('WHITE')
    red = Color('RED')
    blue = Color('BLUE')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.polygon([(10, 10),
                          (40, 25),
                          (10, 40)])
            draw.draw(img)
            assert img[10, 25] == red
            assert img[25, 25] == blue
            assert img[35, 15] == img[35, 35] == white


def test_draw_polyline(fx_wand):
    white = Color('WHITE')
    red = Color('RED')
    blue = Color('BLUE')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.polyline([(10, 10), (40, 25), (10, 40)])
            draw.draw(img)
            assert img[10, 25] == img[25, 25] == blue
            assert img[35, 15] == img[35, 35] == white


def test_draw_push_pop():
    with Drawing() as draw:
        draw.stroke_width = 2
        draw.push()
        draw.stroke_width = 3
        assert 3 == draw.stroke_width
        draw.pop()
        assert 2 == draw.stroke_width


def test_draw_bezier(fx_wand):
    white = Color('WHITE')
    red = Color('RED')
    blue = Color('BLUE')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.bezier([(10, 10),
                         (10, 40),
                         (40, 10),
                         (40, 40)])
            draw.draw(img)
            assert img[10, 10] == img[25, 25] == img[40, 40] == red
            assert img[34, 32] == img[15, 18] == blue
            assert img[34, 38] == img[15, 12] == white


def test_path_curve():
    white = Color('WHITE')
    red = Color('RED')
    blue = Color('BLUE')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.path_start()
            draw.path_move(to=(0, 25), relative=True)
            draw.path_curve(to=(25, 25),
                            controls=((0, 0), (25, 0)))
            draw.path_curve(to=(25, 0),
                            controls=((0, 25), (25, 25)),
                            relative=True)
            draw.path_finish()
            draw.draw(img)
            assert img[25, 25] == red
            assert img[35, 35] == img[35, 35] == blue
            assert img[35, 15] == img[15, 35] == white


def test_path_curve_user_error():
    with Drawing() as draw:
        with raises(TypeError):
            draw.path_curve(to=(5, 7))
        with raises(TypeError):
            draw.path_curve(controls=(5, 7))


def test_path_curve_to_quadratic_bezier():
    white = Color('WHITE')
    red = Color('RED')
    blue = Color('BLUE')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.path_start()
            draw.path_move(to=(0, 25), relative=True)
            draw.path_curve_to_quadratic_bezier(to=(50, 25),
                                                control=(25, 50))
            draw.path_curve_to_quadratic_bezier(to=(-20, -20),
                                                control=(-25, 0),
                                                relative=True)
            draw.path_finish()
            draw.draw(img)
            assert img[30, 5] == red


def test_path_curve_to_quadratic_bezier_smooth():
    white = Color('WHITE')
    red = Color('RED')
    blue = Color('BLUE')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.path_start()
            draw.path_curve_to_quadratic_bezier(to=(25, 25),
                                                control=(25, 25))
            draw.path_curve_to_quadratic_bezier(to=(10, -10),
                                                smooth=True,
                                                relative=True)
            draw.path_curve_to_quadratic_bezier(to=(35, 35),
                                                smooth=True,
                                                relative=False)
            draw.path_curve_to_quadratic_bezier(to=(-10, -10),
                                                smooth=True,
                                                relative=True)
            draw.path_finish()
            draw.draw(img)
            assert img[25, 25] == red
            assert img[30, 30] == blue


def test_path_curve_quadratic_bezier_user_error():
    with Drawing() as draw:
        with raises(TypeError):
            draw.path_curve_to_quadratic_bezier()
        with raises(TypeError):
            draw.path_curve_to_quadratic_bezier(to=(5, 6))


def test_draw_path_elliptic_arc():
    white = Color('WHITE')
    red = Color('RED')
    blue = Color('BLUE')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.path_start()
            draw.path_move(to=(25, 0))
            draw.path_elliptic_arc(to=(25, 50), radius=(15, 25))
            draw.path_elliptic_arc(to=(0, -15), radius=(5, 5),
                                   clockwise=False, relative=True)
            draw.path_close()
            draw.path_finish()
            draw.draw(img)
            assert img[25, 35] == img[25, 20] == red
            assert img[15, 25] == img[30, 45] == blue


def test_draw_path_elliptic_arc_user_error():
    with Drawing() as draw:
        with raises(TypeError):
            draw.path_elliptic_arc(to=(5, 7))
        with raises(TypeError):
            draw.path_elliptic_arc(radius=(5, 7))


def test_draw_path_line():
    white = Color('#FFF')
    red = Color('#F00')
    blue = Color('#00F')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.stroke_width = 10
            draw.path_start()
            draw.path_move(to=(10, 10))
            draw.path_line(to=(40, 40))
            draw.path_line(to=(0, -10), relative=True)
            draw.path_horizontal_line(x=45)
            draw.path_vertical_line(y=25)
            draw.path_horizontal_line(x=-5, relative=True)
            draw.path_vertical_line(y=-5, relative=True)
            draw.path_close()
            draw.path_finish()
            draw.draw(img)
        assert img[40, 40] == img[40, 30] == red
        assert img[45, 25] == img[40, 20] == red


def test_draw_path_line_user_error():
    with Drawing() as draw:
        # Test missing value
        with raises(TypeError):
            draw.path_line()
        with raises(TypeError):
            draw.path_horizontal_line()
        with raises(TypeError):
            draw.path_vertical_line()


def test_draw_move_user_error():
    with Drawing() as draw:
        # Test missing value
        with raises(TypeError):
            draw.path_move()


@mark.parametrize('kwargs', itertools.product(
    [('right', 40), ('width', 30)],
    [('bottom', 40), ('height', 30)]
))
def test_draw_rectangle(kwargs, fx_wand):
    white = Color('WHITE')
    black = Color('BLACK')
    gray = Color('#ccc')
    with Image(width=50, height=50, background=white) as img:
        fx_wand.stroke_width = 2
        fx_wand.fill_color = black
        fx_wand.stroke_color = gray
        fx_wand.rectangle(left=10, top=10, **dict(kwargs))
        fx_wand.draw(img)
        assert (img[7, 7] == img[7, 42] == img[42, 7] ==
                img[42, 42] == img[0, 0] == img[49, 49] == white)
        assert (img[12, 12] == img[12, 38] == img[38, 12] ==
                img[38, 38] == black)


@mark.parametrize('kwargs', itertools.product(
    [('xradius', 10), ('yradius', 10)],
    [('xradius', 20)],
    [('yradius', 20)],
    [('radius', 10)]
))
def test_draw_rectangle_with_radius(kwargs, fx_wand):
    white = Color('WHITE')
    black = Color('BLACK')
    gray = Color('#ccc')
    with Image(width=50, height=50, background=white) as img:
        fx_wand.stroke_width = 2
        fx_wand.fill_color = black
        fx_wand.stroke_color = gray
        fx_wand.rectangle(left=10, top=10,
                          width=30, height=30, **dict(kwargs))
        fx_wand.draw(img)
        assert img[10, 10] == img[40, 40] == white
        assert img[26, 12] == img[26, 36] == black


def test_draw_rotate():
    white = Color('WHITE')
    black = Color('BLACK')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.stroke_color = black
            draw.rotate(45)
            draw.line((3, 3), (35, 35))
            draw.draw(img)
            assert img[0, 49] == black


def test_draw_scale(fx_wand):
    white = Color('WHITE')
    black = Color('BLACK')
    with Image(width=50, height=50, background=white) as img:
        fx_wand.fill_color = black
        fx_wand.scale(x=2.0, y=0.5)
        fx_wand.rectangle(top=5, left=5, width=20, height=20)
        fx_wand.draw(img)
        # if width was scaled up by 200%
        assert img[45, 10] == black
        # if height was scaled down by 50%
        assert img[20, 20] == white


def test_set_fill_pattern_url(fx_wand):
    white = Color('#fff')
    green = Color('#0f0')
    black = Color('#000')
    with Image(width=50, height=50, background=white) as img:
        fx_wand.push_pattern('green_circle', 0, 0, 10, 10)
        fx_wand.fill_color = green
        fx_wand.stroke_color = black
        fx_wand.circle(origin=(5, 5), perimeter=(5, 0))
        fx_wand.pop_pattern()
        fx_wand.set_fill_pattern_url('#green_circle')
        fx_wand.rectangle(top=5, left=5, width=40, height=40)
        fx_wand.draw(img)
        assert img[17, 17] == green


def test_set_stroke_pattern_url(fx_wand):
    white = Color('#fff')
    green = Color('#0f0')
    with Image(width=50, height=50, background=white) as img:
        fx_wand.push_pattern('green_ring', 0, 0, 6, 6)
        fx_wand.fill_color = green
        fx_wand.stroke_color = white
        fx_wand.circle(origin=(3, 3), perimeter=(3, 0))
        fx_wand.pop_pattern()
        fx_wand.set_stroke_pattern_url('#green_ring')
        fx_wand.stroke_width = 6
        fx_wand.rectangle(top=5, left=5, width=40, height=40)
        fx_wand.draw(img)
        assert img[45, 45] == green


def test_draw_skew():
    white = Color('#fff')
    black = Color('#000')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.stroke_color = black
            draw.skew(x=11, y=-24)
            draw.line((3, 3), (35, 35))
            draw.draw(img)
            assert img[43, 42] == black


def test_draw_translate():
    white = Color('#fff')
    black = Color('#000')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.stroke_color = black
            draw.translate(x=5, y=5)
            draw.line((3, 3), (35, 35))
            draw.draw(img)
            assert img[40, 40] == black


def test_draw_text(fx_asset):
    white = Color('#fff')
    with Image(width=100, height=100, background=white) as img:
        was = img.signature
        with Drawing() as draw:
            draw.font = str(fx_asset.join('League_Gothic.otf'))
            draw.font_size = 25
            with Color('#000') as bk:
                draw.fill_color = bk
            draw.gravity = 'west'
            draw.text(0, 0, 'Hello Wand')
            draw.draw(img)
        assert was != img.signature


def test_get_font_metrics_test(fx_asset):
    with Image(width=144, height=192, background=Color('#fff')) as img:
        with Drawing() as draw:
            draw.font = str(fx_asset.join('League_Gothic.otf'))
            draw.font_size = 13
            nm1 = draw.get_font_metrics(img, 'asdf1234')
            nm2 = draw.get_font_metrics(img, 'asdf1234asdf1234')
            nm3 = draw.get_font_metrics(img, 'asdf1234\nasdf1234')
            assert nm1.character_width == draw.font_size
            assert nm1.text_width < nm2.text_width
            assert nm2.text_width <= nm3.text_width
            assert nm2.text_height == nm3.text_height
            m1 = draw.get_font_metrics(img, 'asdf1234', True)
            m2 = draw.get_font_metrics(img, 'asdf1234asdf1234', True)
            m3 = draw.get_font_metrics(img, 'asdf1234\nasdf1234', True)
            assert m1.character_width == draw.font_size
            assert m1.text_width < m2.text_width
            assert m2.text_width > m3.text_width
            assert m2.text_height < m3.text_height


def test_viewbox(fx_asset):
    with Drawing() as draw:
        with raises(TypeError):
            draw.viewbox(None, None, None, None)
        with raises(TypeError):
            draw.viewbox(10, None, None, None)
        with raises(TypeError):
            draw.viewbox(10, 10, None, None)
        with raises(TypeError):
            draw.viewbox(10, 10, 100, None)
        draw.viewbox(10, 10, 100, 100)


def test_regression_issue_163(tmpdir):
    """https://github.com/emcconville/wand/issues/163"""
    unicode_char = b'\xce\xa6'.decode('utf-8')
    with Drawing() as draw:
        with Image(width=500, height=500) as image:
            draw.font_size = 20
            draw.gravity = 'south_west'
            draw.text(0, 0, unicode_char)
            draw(image)
            image.save(filename=str(tmpdir.join('out.jpg')))


def test_set_get_fill_opacity(fx_wand):
    fx_wand.fill_opacity = 1.0
    assert fx_wand.fill_opacity == 1.0


def test_set_get_fill_opacity_user_error(fx_wand):
    with raises(TypeError):
        fx_wand.fill_opacity = "1.5"


def test_set_get_fill_rule(fx_wand):
    valid = 'evenodd'
    notvalid = 'error'
    invalid = (1, 2)
    fx_wand.fill_rule = valid
    assert fx_wand.fill_rule == valid
    with raises(ValueError):
        fx_wand.fill_rule = notvalid
    with raises(TypeError):
        fx_wand.fill_rule = invalid
    fx_wand.fill_rule = 'undefined'  # reset

@mark.skipif(MAGICK_VERSION_NUMBER < 0x700,
             reason='DrawGetOpacity always returns 1.0')
def test_set_get_opacity(fx_wand):
    assert fx_wand.opacity == 1.0
    fx_wand.push()
    fx_wand.opacity = 0.5
    fx_wand.push()
    fx_wand.opacity = 0.25
    assert 0.24 < fx_wand.opacity < 0.26  # Expect float precision issues
    fx_wand.pop()
    assert 0.49 < fx_wand.opacity < 0.51  # Expect float precision issues
    fx_wand.pop()


def test_set_get_stroke_antialias(fx_wand):
    fx_wand.stroke_antialias = False
    assert not fx_wand.stroke_antialias


def test_set_get_stroke_dash_array(fx_wand):
    dash_array = [2, 1, 4, 1]
    fx_wand.stroke_dash_array = dash_array
    assert fx_wand.stroke_dash_array == dash_array


def test_set_get_stroke_dash_offset(fx_wand):
    fx_wand.stroke_dash_offset = 0.5
    assert fx_wand.stroke_dash_offset == 0.5


def test_set_get_stroke_line_cap(fx_wand):
    fx_wand.stroke_line_cap = 'round'
    assert fx_wand.stroke_line_cap == 'round'


def test_set_get_stroke_line_cap_user_error(fx_wand):
    with raises(TypeError):
        fx_wand.stroke_line_cap = 0x74321870
    with raises(ValueError):
        fx_wand.stroke_line_cap = 'apples'


def test_set_get_stroke_line_join(fx_wand):
    fx_wand.stroke_line_join = 'miter'
    assert fx_wand.stroke_line_join == 'miter'


def test_set_get_stroke_line_join_user_error(fx_wand):
    with raises(TypeError):
        fx_wand.stroke_line_join = 0x74321870
    with raises(ValueError):
        fx_wand.stroke_line_join = 'apples'


def test_set_get_stroke_miter_limit(fx_wand):
    fx_wand.stroke_miter_limit = 5
    assert fx_wand.stroke_miter_limit == 5


def test_set_get_stroke_miter_limit_user_error(fx_wand):
    with raises(TypeError):
        fx_wand.stroke_miter_limit = '5'


def test_set_get_stroke_opacity(fx_wand):
    fx_wand.stroke_opacity = 1.0
    assert fx_wand.stroke_opacity == 1.0


def test_set_get_stroke_opacity_user_error(fx_wand):
    with raises(TypeError):
        fx_wand.stroke_opacity = '1.0'


def test_set_get_stroke_width_user_error(fx_wand):
    with raises(TypeError):
        fx_wand.stroke_width = '0.1234'
    with raises(ValueError):
        fx_wand.stroke_width = -1.5


def test_draw_affine(fx_wand):
    skyblue = Color('skyblue')
    black = Color('black')
    with Image(width=100, height=100, background=skyblue) as img:
        img.format = 'png'
        fx_wand.affine([1.5, 0.5, 0, 1.5, 45, 25])
        fx_wand.rectangle(top=5, left=5, width=25, height=25)
        fx_wand.draw(img)
        assert img[25, 25] == skyblue
        assert img[75, 75] == black
        with raises(ValueError):
            fx_wand.affine([1.0])
        with raises(TypeError):
            fx_wand.affine(['a', 'b', 'c', 'd', 'e', 'f'])


def test_draw_clip_path(fx_wand):
    skyblue = Color('skyblue')
    orange = Color('orange')
    with Image(width=100, height=100, background=skyblue) as img:
        fx_wand.push_defs()
        fx_wand.push_clip_path("eyes_only")
        fx_wand.push()
        fx_wand.rectangle(top=0, left=0, width=50, height=50)
        fx_wand.pop()
        fx_wand.pop_clip_path()
        fx_wand.pop_defs()

        fx_wand.clip_path = "eyes_only"
        fx_wand.clip_rule = "nonzero"
        fx_wand.clip_path_units = "object_bounding_box"
        fx_wand.fill_color = orange
        fx_wand.rectangle(top=5, left=5, width=90, height=90)
        fx_wand.draw(img)
        assert img[75, 75] == skyblue
